# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

from numbers import Integral
from warnings import warn

import numpy as np
import pandas as pd
from numpy import dot, hstack
from numpy.linalg import qr, svd
from scipy.linalg import eigh

from skbio.util import get_rng
from skbio.stats.distance import DistanceMatrix
from ._ordination_results import OrdinationResults
from ._utils import center_distance_matrix, scale


def pcoa(
    distance_matrix,
    method="eigh",
    number_of_dimensions=0,
    inplace=False,
    seed=None,
    warn_neg_eigval=0.01,
):
    r"""Perform Principal Coordinate Analysis (PCoA).

    PCoA is an ordination method similar to Principal Components Analysis (PCA), with
    the difference that it operates on distance matrices, calculated using meaningful
    and typically non-Euclidian methods.

    Parameters
    ----------
    distance_matrix : DistanceMatrix
        The input distance matrix.
    method : str, optional
        Matrix decomposition method to use. Default is "eigh" (eigendecomposition),
        which computes exact eigenvectors and eigenvalues for all dimensions. The
        alternate is "fsvd" (fast singular value decomposition), a heuristic that can
        compute only a given number of dimensions.
    number_of_dimensions : int or float, optional
        Dimensions to reduce the distance matrix to. This number determines how many
        eigenvectors and eigenvalues will be returned. If an integer is provided, the
        exact number of dimensions will be retained. If a float between 0 and 1, it
        represents the fractional cumulative variance to be retained. Default is 0,
        which will retain the same number of dimensions as the distance matrix.
    inplace : bool, optional
        If True, the input distance matrix will be centered in-place to reduce memory
        consumption, at the cost of losing the original distances. Default is False.
    seed : int or np.random.Generator, optional
        A user-provided random seed or random generator instance for method "fsvd".
        See :func:`details <skbio.util.get_rng>`.

        .. versionadded:: 0.6.3

    warn_neg_eigval : bool or float, optional
        Raise a warning if any negative eigenvalue is obtained and its magnitude
        exceeds the specified fraction threshold compared to the largest positive
        eigenvalue, which suggests potential inaccuracy in the PCoA result. Default is
        0.01. Set True to warn regardless of the magnitude. Set False to disable
        warning completely.

        .. versionadded:: 0.6.3

    Returns
    -------
    OrdinationResults
        Object that stores the PCoA results, including eigenvalues, the proportion
        explained by each of them, and transformed sample coordinates.

    See Also
    --------
    OrdinationResults

    Notes
    -----
    Principal Coordinate Analysis (PCoA) was first described in [1]_.

    This function uses a choice of two methods for matrix decomposition: The default
    method, ``eigh``, performs eigendecomposition, an exact method that computes all
    eigenvectors and eigenvalues. The alternative method, ``fsvd``, performs fast
    singular value decomposition (FSVD) [2]_, an efficient heuristic method that
    allows a custom number of dimensions to be specified to reduce calculation at the
    cost of losing accuracy. The degree of accuracy lost is dependent on dataset.

    Note that the default method ``eigh`` does not natively support a given number of
    dimensions to reduce a matrix to. Therefore, if this parameter is specified, all
    eigenvectors and eigenvalues will be simply be computed with no speed gain, and
    only the specified number of dimensions will be returned.

    Eigenvalues represent the magnitude of individual principal coordinates, and
    they are usually positive. However, negative eigenvalues can occur when the
    distances were calculated using a non-Euclidean metric that does not satisfy
    triangle inequality. If the negative eigenvalues are small in magnitude compared
    to the largest positive eigenvalue, it is usually safe to ignore them. However,
    large negative eigenvalues may indicate result inaccuracy, in which case a warning
    message will be displayed. The paramter ``warn_neg_eigval`` controls the threshold
    for the warning.

    PCoA on Euclidean distances is equivalent to Principal Component Analysis (PCA).
    However, in ecology, the Euclidean distance preserved by PCA is often not a good
    choice because it deals poorly with double zeros. For example, species have
    unimodal distributions along environmental gradients. If a species is absent from
    two sites simultaneously, it can't be known if an environmental variable is too
    high in one of them and too low in the other, or too low in both, etc. On the other
    hand, if a species is present in two sites, that means that the sites are similar.

    Note that the returned eigenvectors are not normalized to unit length.

    References
    ----------
    .. [1] Gower, J. C. (1966). Some distance properties of latent root and vector
       methods used in multivariate analysis. Biometrika, 53(3-4), 325-338.

    .. [2] Halko, N., Martinsson, P. G., Shkolnisky, Y., & Tygert, M. (2011). An
       algorithm for the principal component analysis of large data sets. SIAM
       Journal on Scientific computing, 33(5), 2580-2594.

    """
    distance_matrix = DistanceMatrix(distance_matrix)

    # Center distance matrix, a requirement for PCoA here
    matrix_data = center_distance_matrix(distance_matrix.data, inplace=inplace)

    # If no dimension specified, by default will compute all eigenvectors
    # and eigenvalues
    if number_of_dimensions == 0:
        if method == "fsvd" and matrix_data.shape[0] > 10:
            warn(
                "FSVD: since no value for number_of_dimensions is specified, "
                "PCoA for all dimensions will be computed, which may "
                "result in long computation time if the original "
                "distance matrix is large.",
                RuntimeWarning,
            )

        # distance_matrix is guaranteed to be square
        number_of_dimensions = matrix_data.shape[0]
    elif number_of_dimensions < 0:
        raise ValueError(
            "Invalid operation: cannot reduce distance matrix "
            "to negative dimensions using PCoA. Did you intend "
            'to specify the default value "0", which sets '
            "the number_of_dimensions equal to the "
            "dimensionality of the given distance matrix?"
        )
    elif not isinstance(number_of_dimensions, Integral) and number_of_dimensions > 1:
        raise ValueError(
            "Invalid operation: A floating-point number greater than 1 cannot be "
            "supplied as the number of dimensions."
        )

    if warn_neg_eigval and not 0 <= warn_neg_eigval <= 1:
        raise ValueError(
            "warn_neg_eigval must be Boolean or a floating-point number between 0 "
            "and 1."
        )

    # Perform eigendecomposition
    if method == "eigh":
        # eigh does not natively support specifying number_of_dimensions, i.e.
        # there are no speed gains unlike in FSVD. Later, we slice off unwanted
        # dimensions to conform the result of eigh to the specified
        # number_of_dimensions.

        eigvals, eigvecs = eigh(matrix_data)
        long_method_name = "Principal Coordinate Analysis"
    elif method == "fsvd":
        # new parameter for num_dimensions = number of dimensions (accounting for
        # non-int values)
        num_dimensions = number_of_dimensions
        if 0 < number_of_dimensions < 1:
            warn(
                "FSVD: since value for number_of_dimensions is specified as float, "
                "PCoA for all dimensions will be computed, which may "
                "result in long computation time if the original "
                "distance matrix is large. "
                "Consider specifying an integer value to optimize performance.",
                RuntimeWarning,
            )
            num_dimensions = matrix_data.shape[0]
        eigvals, eigvecs = _fsvd(matrix_data, num_dimensions, seed=seed)
        long_method_name = "Approximate Principal Coordinate Analysis using FSVD"
    else:
        raise ValueError(
            "PCoA eigendecomposition method {} not supported.".format(method)
        )
    # Ensure number_of_dimensions does not exceed available dimensions
    # number_of_dimensions = min(number_of_dimensions, eigvals.shape[0])

    # cogent makes eigenvalues positive by taking the
    # abs value, but that doesn't seem to be an approach accepted
    # by L&L to deal with negative eigenvalues. We raise a warning
    # in that case. First, we make values close to 0 equal to 0.
    negative_close_to_zero = np.isclose(eigvals, 0)
    eigvals[negative_close_to_zero] = 0

    # eigvals might not be ordered, so we first sort them, then analogously
    # sort the eigenvectors by the ordering of the eigenvalues too
    idxs_descending = eigvals.argsort()[::-1]
    eigvals = eigvals[idxs_descending]
    eigvecs = eigvecs[:, idxs_descending]

    # large negative eigenvalues suggest result inaccuracy
    # see: https://github.com/scikit-bio/scikit-bio/issues/1410
    if warn_neg_eigval and eigvals[-1] < 0:
        if warn_neg_eigval is True or -eigvals[-1] > eigvals[0] * warn_neg_eigval:
            warn(
                "The result contains negative eigenvalues that are large in magnitude,"
                " which may suggest result inaccuracy. See Notes for details. The"
                " negative-most eigenvalue is {0} whereas the largest positive one is"
                " {1}.".format(eigvals[-1], eigvals[0]),
                RuntimeWarning,
            )

    # If we return only the coordinates that make sense (i.e., that have a
    # corresponding positive eigenvalue), then Jackknifed Beta Diversity
    # won't work as it expects all the OrdinationResults to have the same
    # number of coordinates. In order to solve this issue, we return the
    # coordinates that have a negative eigenvalue as 0
    num_positive = (eigvals >= 0).sum()
    eigvecs[:, num_positive:] = np.zeros(eigvecs[:, num_positive:].shape)
    eigvals[num_positive:] = np.zeros(eigvals[num_positive:].shape)

    if method == "fsvd":
        # Since the dimension parameter, hereafter referred to as 'd',
        # restricts the number of eigenvalues and eigenvectors that FSVD
        # computes, we need to use an alternative method to compute the sum
        # of all eigenvalues, used to compute the array of proportions
        # explained. Otherwise, the proportions calculated will only be
        # relative to d number of dimensions computed; whereas we want
        # it to be relative to the entire dimensionality of the
        # centered distance matrix.

        # An alternative method of calculating th sum of eigenvalues is by
        # computing the trace of the centered distance matrix.
        # See proof outlined here: https://goo.gl/VAYiXx
        sum_eigenvalues = np.trace(matrix_data)
    else:
        # Calculate proportions the usual way
        sum_eigenvalues = np.sum(eigvals)

    proportion_explained = eigvals / sum_eigenvalues
    if 0 < number_of_dimensions < 1:
        cumulative_variance = np.cumsum(proportion_explained)
        num_dimensions = (
            np.searchsorted(cumulative_variance, number_of_dimensions, side="left") + 1
        )
        # gives the number of dimensions needed to reach specified variance
        # updates number of dimensions to reach the requirement of variance.
        number_of_dimensions = num_dimensions

    # In case eigh is used, eigh computes all eigenvectors and -values.
    # So if number_of_dimensions was specified, we manually need to ensure
    # only the requested number of dimensions
    # (number of eigenvectors and eigenvalues, respectively) are returned.
    eigvecs = eigvecs[:, :number_of_dimensions]
    eigvals = eigvals[:number_of_dimensions]
    proportion_explained = proportion_explained[:number_of_dimensions]

    # Scale eigenvalues to have length = sqrt(eigenvalue). This
    # works because np.linalg.eigh returns normalized
    # eigenvectors. Each row contains the coordinates of the
    # objects in the space of principal coordinates. Note that at
    # least one eigenvalue is zero because only n-1 axes are
    # needed to represent n points in a euclidean space.
    coordinates = eigvecs * np.sqrt(eigvals)

    axis_labels = ["PC%d" % i for i in range(1, number_of_dimensions + 1)]
    return OrdinationResults(
        short_method_name="PCoA",
        long_method_name=long_method_name,
        eigvals=pd.Series(eigvals, index=axis_labels),
        samples=pd.DataFrame(
            coordinates, index=distance_matrix.ids, columns=axis_labels
        ),
        proportion_explained=pd.Series(proportion_explained, index=axis_labels),
    )


def _fsvd(centered_distance_matrix, number_of_dimensions=10, seed=None):
    """Perform singular value decomposition.

    More specifically in this case eigendecomposition, using fast heuristic algorithm
    nicknamed "FSVD" (FastSVD), adapted and optimized from the algorithm described
    by Halko et al (2011).

    Parameters
    ----------
    centered_distance_matrix : np.array
       Numpy matrix representing the distance matrix for which the
       eigenvectors and eigenvalues shall be computed
    number_of_dimensions : int
       Number of dimensions to keep. Must be lower than or equal to the
       rank of the given distance_matrix.
    seed : int or np.random.Generator, optional
        A user-provided random seed or random generator instance.

    Returns
    -------
    np.array
       Array of eigenvectors, each with number_of_dimensions length.
    np.array
       Array of eigenvalues, a total number of number_of_dimensions.

    Notes
    -----
    The algorithm is based on [1]_.

    Ported from MATLAB implementation described in [2]_.

    References
    ----------
    .. [1] Halko, N., Martinsson, P. G., Shkolnisky, Y., & Tygert, M. (2011). An
       algorithm for the principal component analysis of large data sets. SIAM
       Journal on Scientific computing, 33(5), 2580-2594.

    .. [2] https://stats.stackexchange.com/a/11934/211065

    """
    m, n = centered_distance_matrix.shape

    # Number of levels of the Krylov method to use.
    # For most applications, num_levels=1 or num_levels=2 is sufficient.
    num_levels = 1

    # Changes the power of the spectral norm, thus minimizing the error).
    use_power_method = False

    # Note: a (conjugate) transpose is removed for performance, since we
    # only expect square matrices.
    if m != n:
        raise ValueError("FSVD expects square distance matrix")

    if number_of_dimensions > m or number_of_dimensions > n:
        raise ValueError(
            "FSVD: number_of_dimensions cannot be larger than"
            " the dimensionality of the given distance matrix."
        )

    if number_of_dimensions < 0:
        raise ValueError(
            "Invalid operation: cannot reduce distance matrix "
            "to negative dimensions using PCoA. Did you intend "
            'to specify the default value "0", which sets '
            "the number_of_dimensions equal to the "
            "dimensionality of the given distance matrix?"
        )

    k = number_of_dimensions + 2

    # Form a real nxl matrix G whose entries are independent, identically
    # distributed Gaussian random variables of zero mean and unit variance
    rng = get_rng(seed)
    G = rng.standard_normal(size=(n, k))

    if use_power_method:
        # use only the given exponent
        H = dot(centered_distance_matrix, G)

        for x in range(2, num_levels + 2):
            # enhance decay of singular values
            # note: distance_matrix is no longer transposed, saves work
            # since we're expecting symmetric, square matrices anyway
            # (Daniel McDonald's changes)
            H = dot(centered_distance_matrix, dot(centered_distance_matrix, H))

    else:
        # compute the m x l matrices H^{(0)}, ..., H^{(i)}
        # Note that this is done implicitly in each iteration below.
        H = dot(centered_distance_matrix, G)
        # to enhance performance
        H = hstack((H, dot(centered_distance_matrix, dot(centered_distance_matrix, H))))
        for x in range(3, num_levels + 2):
            tmp = dot(centered_distance_matrix, dot(centered_distance_matrix, H))

            H = hstack(
                (H, dot(centered_distance_matrix, dot(centered_distance_matrix, tmp)))
            )

    # Using the pivoted QR-decomposition, form a real m * ((i+1)l) matrix Q
    # whose columns are orthonormal, s.t. there exists a real
    # ((i+1)l) * ((i+1)l) matrix R for which H = QR
    Q, R = qr(H)

    # Compute the n * ((i+1)l) product matrix T = A^T Q
    T = dot(centered_distance_matrix, Q)  # step 3

    # Form an SVD of T
    Vt, St, W = svd(T, full_matrices=False)
    W = W.transpose()

    # Compute the m * ((i+1)l) product matrix
    Ut = dot(Q, W)

    U_fsvd = Ut[:, :number_of_dimensions]

    S = St[:number_of_dimensions]

    # drop imaginary component, if we got one
    # Note:
    #   In cogent, after computing eigenvalues/vectors, the imaginary part
    #   is dropped, if any. We know for a fact that the eigenvalues are
    #   real, so that's not necessary, but eigenvectors can in principle
    #   be complex (see for example
    #   http://math.stackexchange.com/a/47807/109129 for details)
    eigenvalues = S.real
    eigenvectors = U_fsvd.real

    return eigenvalues, eigenvectors


def pcoa_biplot(ordination, y):
    """Compute the projection of descriptors into a PCoA matrix.

    Parameters
    ----------
    ordination: OrdinationResults
        The computed principal coordinates analysis of dimensions (n, c) where
        the matrix ``y`` will be projected onto.
    y: DataFrame
        Samples by features table of dimensions (n, m). These can be
        environmental features or abundance counts. This table should be
        normalized in cases of dimensionally heterogenous physical variables.

    Returns
    -------
    OrdinationResults
        The modified input object that includes projected features onto the
        ordination space in the ``features`` attribute.

    Notes
    -----
    This implementation is as described in Chapter 9 of [1]_.

    References
    ----------
    .. [1] Legendre P. and Legendre L. 1998. Numerical Ecology. Elsevier, Amsterdam.

    """
    # acknowledge that most saved ordinations lack a name, however if they have
    # a name, it should be PCoA
    if ordination.short_method_name != "" and ordination.short_method_name != "PCoA":
        raise ValueError(
            "This biplot computation can only be performed in a " "PCoA matrix."
        )

    if set(y.index) != set(ordination.samples.index):
        raise ValueError(
            "The eigenvectors and the descriptors must describe " "the same samples."
        )

    eigvals = ordination.eigvals.values
    coordinates = ordination.samples
    N = coordinates.shape[0]

    # align the descriptors and eigenvectors in a sample-wise fashion
    y = y.reindex(coordinates.index)

    # S_pc from equation 9.44
    # Represents the covariance matrix between the features matrix and the
    # column-centered eigenvectors of the pcoa.
    spc = (1 / (N - 1)) * y.values.T.dot(scale(coordinates, ddof=1))

    # U_proj from equation 9.55, is the matrix of descriptors to be projected.
    #
    # Only get the power of non-zero values, otherwise this will raise a
    # divide by zero warning. There shouldn't be negative eigenvalues(?)
    Uproj = np.sqrt(N - 1) * spc.dot(
        np.diag(np.power(eigvals, -0.5, where=eigvals > 0))
    )

    ordination.features = pd.DataFrame(
        data=Uproj, index=y.columns.copy(), columns=coordinates.columns.copy()
    )
    ordination.features.fillna(0.0, inplace=True)

    return ordination



# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

# Script to insert a meta description tag into each documentation page. The
# content is the first line of the docstring of the corresponding object.

import os
import re
import glob
import html


# -- Configuration -----------------------------------------------------------

rootdir = "build/html"
max_length = 160


# -- Workflow ----------------------------------------------------------------

dd_pattern = re.compile(r'<dd>(.*?)</dd>', flags=re.DOTALL)
p_pattern = re.compile(r'<p>(.*?)</p>', flags=re.DOTALL)
meta_pattern = re.compile(r'(\s*)(<meta\b[^>]*\/?>)')

cwd = os.getcwd()
os.chdir(os.path.join(os.path.dirname(__file__), rootdir))

for file in glob.glob("**/*.html", recursive=True):
    with open(file, "r") as fh:
        content = fh.read()

    # find first line of docstring (summary)
    dd_match = dd_pattern.search(content)
    if not dd_match:
        continue
    p_match = p_pattern.search(dd_match.group(1))
    if not p_match:
        continue
    summary = p_match.group(1)

    # truncate summary to given length
    if len(summary) > max_length:
        summary = summary[:max_length - 3]
        if ' ' in summary:
            summary = summary.rsplit(' ', 1)[0]
        summary += "..."

    # make summary safe for HTML
    summary = html.escape(summary)

    # insert summary into metadata
    line = f'<meta name="description" content="{summary}" />'
    meta_match = meta_pattern.search(content)
    if not meta_match:
        continue
    indent = meta_match.group(1)
    pos = meta_match.end()
    content = content[:pos] + indent + line + content[pos:]

    with open(file, "w") as fh:
        fh.write(content)

os.chdir(cwd)



# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

import inspect
from types import FunctionType
from numbers import Integral

import numpy as np


def resolve_key(obj, key):
    """Resolve key given an object and key."""
    if callable(key):
        return key(obj)
    elif hasattr(obj, "metadata"):
        return obj.metadata[key]
    raise TypeError(
        "Could not resolve key %r. Key must be callable or %s must"
        " have `metadata` attribute." % (key, obj.__class__.__name__)
    )


def make_sentinel(name):
    return type(
        name,
        (),
        {"__repr__": lambda s: name, "__str__": lambda s: name, "__class__": None},
    )()


def find_sentinels(function, sentinel):
    params = inspect.signature(function).parameters
    return [name for name, param in params.items() if param.default is sentinel]


class MiniRegistry(dict):
    def __call__(self, name):
        """Act as a decorator to register functions with self."""

        def decorator(func):
            self[name] = func
            return func

        return decorator

    def copy(self):
        """Use for inheritance."""
        return self.__class__(super(MiniRegistry, self).copy())

    def formatted_listing(self):
        """Produce an RST list with descriptions."""
        if len(self) == 0:
            return "\tNone"
        else:
            return "\n".join(
                [
                    "\t%r\n\t  %s" % (name, self[name].__doc__.split("\n")[0])
                    for name in sorted(self)
                ]
            )

    def interpolate(self, obj, name):
        """Inject the formatted listing in the second blank line of `name`."""
        f = getattr(obj, name)
        f2 = FunctionType(
            f.__code__,
            f.__globals__,
            name=f.__name__,
            argdefs=f.__defaults__,
            closure=f.__closure__,
        )

        # Conveniently the original docstring is on f2, not the new ones if
        # inheritance is happening. I have no idea why.
        t = f2.__doc__.split("\n\n")
        t.insert(2, self.formatted_listing())
        f2.__doc__ = "\n\n".join(t)

        setattr(obj, name, f2)


def chunk_str(s, n, char):
    """Insert `char` character every `n` characters in string `s`.

    Canonically pronounced "chunkster".

    """
    # Modified from http://stackoverflow.com/a/312464/3776794
    if n < 1:
        raise ValueError(
            "Cannot split string into chunks with n=%d. n must be >= 1." % n
        )
    return char.join((s[i : i + n] for i in range(0, len(s), n)))


def cardinal_to_ordinal(n):
    """Return ordinal string version of cardinal int `n`.

    Parameters
    ----------
    n : int
        Cardinal to convert to ordinal. Must be >= 0.

    Returns
    -------
    str
        Ordinal version of cardinal `n`.

    Raises
    ------
    ValueError
        If `n` is less than 0.

    Notes
    -----
    This function can be useful when writing human-readable error messages.

    Examples
    --------
    >>> from skbio.util import cardinal_to_ordinal
    >>> cardinal_to_ordinal(0)
    '0th'
    >>> cardinal_to_ordinal(1)
    '1st'
    >>> cardinal_to_ordinal(2)
    '2nd'
    >>> cardinal_to_ordinal(3)
    '3rd'

    """
    # Taken and modified from http://stackoverflow.com/a/20007730/3776794
    # Originally from http://codegolf.stackexchange.com/a/4712 by Gareth
    if n < 0:
        raise ValueError("Cannot convert negative integer %d to ordinal " "string." % n)
    return "%d%s" % (n, "tsnrhtdd"[(n // 10 % 10 != 1) * (n % 10 < 4) * n % 10 :: 4])


def safe_md5(open_file, block_size=2**20):
    """Compute an md5 sum without loading the file into memory.

    Parameters
    ----------
    open_file : file object
        open file handle to the archive to compute the checksum. It
        must be open as a binary file
    block_size : int, optional
        size of the block taken per iteration

    Returns
    -------
    md5 : md5 object from the hashlib module
        object with the loaded file

    Notes
    -----
    This method is based on the answers given in:
    http://stackoverflow.com/a/1131255/379593

    Examples
    --------
    >>> from io import BytesIO
    >>> from skbio.util import safe_md5
    >>> fd = BytesIO(b"foo bar baz") # open file like object
    >>> x = safe_md5(fd)
    >>> x.hexdigest()
    'ab07acbb1e496801937adfa772424bf7'
    >>> fd.close()

    """
    import hashlib

    md5 = hashlib.md5()
    data = True
    while data:
        data = open_file.read(block_size)
        if data:
            md5.update(data)
    return md5


def find_duplicates(iterable):
    """Find duplicate elements in an iterable.

    Parameters
    ----------
    iterable : iterable
        Iterable to be searched for duplicates (i.e., elements that are
        repeated).

    Returns
    -------
    set
        Repeated elements in `iterable`.

    """
    # modified from qiita.qiita_db.util.find_repeated
    # https://github.com/biocore/qiita
    # see licenses/qiita.txt
    seen, repeated = set(), set()
    for e in iterable:
        if e in seen:
            repeated.add(e)
        else:
            seen.add(e)
    return repeated


def get_rng(seed=None):
    r"""Get a random generator.

    .. versionchanged:: 0.6.3
        Added legacy random generator support.

    Parameters
    ----------
    seed : int, Generator or RandomState, optional
        A user-provided random seed or random generator instance.

    Returns
    -------
    np.random.Generator
        Random generator instance.

    See Also
    --------
    numpy.random.Generator
    numpy.random.RandomState

    Notes
    -----
    A random generator ensures reproducibility of outputs. scikit-bio utilizes NumPy's
    new random generator (``Generator``) [1]_, which was introduced in version 1.17.
    See NEP 19 [3]_ for an introduction to this change.

    The following code demonstrates the recommended usage of the random generator with
    various scikit-bio functions that are stochastic. With a random generator created
    in advance, you can plug it into multiple function calls. The results of the entire
    code will be reproducible.

    (42 is an arbitrarily chosen random seed. It can be any non-negative integer.)

    .. code-block:: python

       rng = np.random.default_rng(42)
       skbio_func1(..., seed=rng)
       skbio_func2(..., seed=rng)
       ...

    Alternatively, you may specify an integer seed to make the result of each function
    call reproducible, such as:

    .. code-block:: python

       skbio_func1(..., seed=42)
       skbio_func2(..., seed=42)
       ...

    Meanwhile, scikit-bio respects the legacy random generator (``RandomState``) [4]_.
    If ``np.random.seed`` has been called, or a ``RandomState`` instance is provided,
    scikit-bio will create a new random generator from a seed selected by the legacy
    random generator. This ensures reproducibility of legacy code and compatibility
    with packages that use the legacy mechanism. For example:

    .. code-block:: python

       np.random.seed(42)
       skbio_func1(...)
       skbio_func2(...)
       ...

    It should be noted that the legacy random generator will not be directly used by
    scikit-bio functions for random number generation. Only the new random generator
    will be exposed to the functions.

    References
    ----------
    .. [1] https://numpy.org/devdocs/reference/random/generator.html

    .. [2] https://numpy.org/doc/stable/reference/random/legacy.html

    .. [3] https://numpy.org/neps/nep-0019-rng-policy.html

    .. [4] https://numpy.org/doc/stable/reference/random/legacy.html

    """
    # seed is new Generator: directly return it
    try:
        if isinstance(seed, np.random.Generator):
            return seed
    except AttributeError:
        raise ValueError(
            "The installed NumPy version does not support random.Generator. "
            "Please use NumPy >= 1.17."
        )

    # no seed: create a generator from a seed selected by the legacy random state
    # this ensures reproducibility of legacy code starting with `np.random.seed(x)`
    if seed is None:
        # the result of `np.iinfo(np.int32).max + 1` is usually 2,147,483,648, or
        # 1 << 31, # which is also the result of `np.random.get_state()[1][0]` if
        # `np.random` hasn't been called
        seed = np.random.randint(np.iinfo(np.int32).max + 1)

    # seed is legacy RandomState: create a generator from a seed selected by it
    elif isinstance(seed, np.random.RandomState):
        seed = seed.randint(np.iinfo(np.int32).max + 1)

    # seed is integer: create a generator from this random state
    elif not isinstance(seed, Integral):
        raise ValueError(
            "Invalid seed. It must be an integer or a random generator instance."
        )

    return np.random.default_rng(seed)



# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

import numpy as np

from skbio.tree import TreeNode
from skbio.util._decorator import params_aliased
from skbio.util._warning import _warn_deprecated
from ._c_nj import nj_minq_cy
from ._utils import _check_dm


@params_aliased([("neg_as_zero", "disallow_negative_branch_length", "0.6.3", True)])
def nj(
    dm,
    neg_as_zero=True,
    result_constructor=None,
    inplace=False,
):
    r"""Perform neighbor joining (NJ) for phylogenetic reconstruction.

    Parameters
    ----------
    dm : skbio.DistanceMatrix
        Input distance matrix containing pairwise distances among taxa.
    neg_as_zero : bool, optional
        If True (default), convert negative branch lengths into zeros.
    result_constructor : function, optional
        Function to apply to construct the result object. This must take a
        newick-formatted string as input. Deprecated and to be removed in a
        future release.

        .. deprecated:: 0.6.3

    inplace : bool, optional
        If True, the input distance matrix will be manipulated in-place to reduce
        memory consumption, at the cost of losing the original data. Default is False.

        .. versionadded:: 0.6.3

    Returns
    -------
    TreeNode
        Reconstructed phylogenetic tree.

        .. versionchanged:: 0.6.3
            The NJ algorithm has been optimized. The output may be slightly different
            from the previous version in root placement, node ordering, and the numeric
            precision of branch lengths. However, the overall tree topology and branch
            lengths should remain the same.

    See Also
    --------
    bme
    upgma
    TreeNode.root_at_midpoint

    Notes
    -----
    Neighbor joining (NJ) was initially described by Saitou and Nei (1987) [1]_. It is
    a simple and efficient agglomerative clustering method that builds a phylogenetic
    tree based on a distance matrix.

    This function implements the canonical NJ method. The algorithm has been optimized
    to improve efficiency, but no heuristic was involved to further accelerate it at
    the cost of accuracy. Therefore, one is guaranteed to obtain an optimal tree. The
    algorithm is *O*\(*n*:sup:`3`) in time and *O*\(*n*:sup:`2`) in space.

    NJ creates an **unrooted** tree with varying tip heights. This contrasts UPGMA
    (:func:`upgma`), which always produces ultrametric trees. One may subsequently use
    choice of strategies such as midpoint rooting (:meth:`~TreeNode.root_at_midpoint`)
    or outgroup rooting (:meth:`~TreeNode.root_by_outgroup`) to convert the result into
    a rooted tree.

    NJ is most accurate when distances are **additive** -- the distance between two
    taxa in the matrix equals to the sum of branch lengths connecting them in the tree.
    When this assumption is violated, which is common in real studies, negative branch
    lengths may be produced, challenging interpretation and subsequent analyses. To
    address this issue, this function converts negative branch lengths into zeros by
    default, but this behavior can be disabled by setting ``neg_as_zero`` to False.

    Gascuel and Steel (2006) provide a detailed overview of neighbor joining in terms
    of its biological relevance and limitations [2]_. They proved that NJ is a greedy
    heuristic to the balanced minimum evolution (BME) problem. An alternative method
    to this problem is provided by :func:`bme`.

    The example presented here is derived from the Wikipedia page on neighbor joining
    [3]_.

    References
    ----------
    .. [1] Saitou, N., & Nei, M. (1987). The neighbor-joining method: a new method for
       reconstructing phylogenetic trees. Mol Biol Evol, 4(4), 406-425.

    .. [2] Gascuel, O., & Steel, M. (2006). Neighbor-joining revealed. Mol Biol Evol,
       23(11), 1997-2000.

    .. [3] http://en.wikipedia.org/wiki/Neighbour_joining

    Examples
    --------
    Define a new distance matrix object describing the distances between five
    taxa: a, b, c, d, and e.

    >>> from skbio import DistanceMatrix
    >>> from skbio.tree import nj

    >>> data = [[0,  5,  9,  9,  8],
    ...         [5,  0, 10, 10,  9],
    ...         [9, 10,  0,  8,  7],
    ...         [9, 10,  8,  0,  3],
    ...         [8,  9,  7,  3,  0]]
    >>> ids = list('abcde')
    >>> dm = DistanceMatrix(data, ids)

    Construct the neighbor joining tree representing the relationship between
    those taxa. This is returned as a TreeNode object.

    >>> tree = nj(dm)
    >>> print(tree.ascii_art())
                                  /-a
                        /--------|
              /--------|          \-b
             |         |
             |          \-c
    ---------|
             |--e
             |
              \-d

    """
    if result_constructor is not None:
        msg = (
            "`result_constructor` is deprecated and will be removed in a future "
            "release."
        )
        _warn_deprecated(nj, "0.6.3", msg)

    _check_dm(dm)
    taxa = list(dm.ids)

    dm_ = dm.data
    if not inplace:
        dm_ = dm_.copy()

    lm = _nj(dm_)

    tree = _tree_from_linkmat(lm, taxa, rooted=False, neg_as_zero=neg_as_zero)
    if result_constructor is not None:
        tree = result_constructor(str(tree))
    return tree


def _nj(dm):
    r"""Perform neighbor joining (NJ) for phylogenetic reconstruction.

    Parameters
    ----------
    dm : (N, N) ndarray
        Input distance matrix.

    Returns
    -------
    ndarray of shape (N - 1, 4)
        Linkage matrix representing the tree.

    Notes
    -----
    This function manipulates the distance matrix in-place. Therefore, one should make
    a copy prior to running the function if the original distance matrix needs to be
    preserved.

    """
    # This function re-uses the original array space during iteration, without creating
    # additional intermediate arrays. Therefore, it is memory-efficient, and avoids the
    # time overhead of allocating new memory space.

    # This function only operates on arrays of numbers, therefore it can be further
    # Cythonized. However, Cythonization did not bring significant performance gain in
    # tests, likely because all operations already utilize NumPy APIs. That being said,
    # further optimization and testing should be convenient.

    N = n = dm.shape[0]  # dimension
    sums = dm.sum(axis=0)  # distance sums
    idxs = np.arange(N)  # cluster indices
    lm = np.empty((N - 1, 4))  # linkage matrix

    # Iteratively merge taxa until there are three left.
    while n > 3:
        # Create memory views of currently relevant array areas.
        dm_ = dm[:n, :n]
        sums_ = sums[:n]
        idxs_ = idxs[:n]

        # Find the minimum value of the Q-matrix and return its position (i, j).
        #   Q(i, j) = (n - 2) d(i, j) - \sum d(i) - \sum d(j)
        # The function call avoids constructing the entire Q-matrix, but instead
        # computes values and finds the minimum as the computation goes.
        i, j = nj_minq_cy(dm_, sums_)

        # Get half of the original distance at (i, j).
        d_ij_ = dm[i, j] / 2

        # Taxa i and j will be merged into a cluster {i, j}. The updated distance from
        # cluster to any other taxon k is:
        #   d({i, j}, k) = (d(i, k) + d(j, k) - d(i, j)) / 2
        # We first compute (d(i, k) + d(j, k)) / 2 and save the results in row i.
        dm_[i] += dm_[j]
        dm_[i] /= 2

        # Compute branch lengths of taxa i and j.
        #   \delta = (\sum d(i) - \sum d(j)) / (2(n - 2))
        #   L(i) = d(i, j) / 2 + \delta
        #   L(j) = d(i, j) / 2 - \delta
        delta_ = (sums_[i] - sums_[j]) / (2 * n - 4)
        L_i = d_ij_ + delta_
        L_j = d_ij_ - delta_

        # The previously calculated sums can be updated for re-use. Specifically, for
        # taxon k, there is:
        #   new sum = old sum - d(i, k) - d(j, k) + d({i,j}, k)
        #           = old sum - d(i, k) - d(j, k) + (d(i, k) + d(j, k) - d(i, j)) / 2
        #           = old sum - (d(i, k) + d(j, k)) / 2 - d(i, j) / 2
        # We already have (d(i, k) + d(j, k)) / 2 stored in row i, therefore:
        sums_[:] -= dm_[i]
        sums_[:] -= d_ij_

        # Now complete the calculation of the updated distances d({i, j}, k).
        dm_[i] -= d_ij_

        # Update column i to match row i.
        dm_[:, i] = dm_[i]

        # Because two taxa have been merged into one cluster, we will shrink the
        # distance matrix from (n, n) to (n - 1, n - 1). Specifically, we will move
        # the last row/column (index: n - 1) to row/column j.
        n_1 = n - 1
        dm_[j] = dm_[n_1]
        dm_[:, j] = dm_[:, n_1]

        # Also move the last sum to j.
        sums_[j] = sums_[n_1]

        # Then calculate the updated sum at i (now cluster {i, j}), which is the sum
        # of the updated distances.
        sums_[i] = dm_[i, :n_1].sum()

        # Store the taxa and branch lengths to the linkage matrix.
        lm[N - n] = idxs_[i], idxs_[j], L_i, L_j

        # Update cluster indices. Specifically, position i will have the new cluster
        # index. Meanwhile, position j will be replaced with the last cluster.
        idxs_[i] = 2 * N - n
        idxs_[j] = idxs_[n_1]

        n -= 1

    # Perform final calculation on the three remaining taxa. They will become children
    # of the root node, and the entire tree is unrooted.
    L_0 = (dm[0, 1] + dm[0, 2] - dm[1, 2]) / 2
    lm[N - 3] = idxs[1], idxs[2], dm[0, 1] - L_0, dm[0, 2] - L_0
    lm[N - 2] = idxs[0], 2 * N - 3, L_0, 0

    return lm


def _tree_from_linkmat(lm, taxa, rooted=True, neg_as_zero=True):
    r"""Convert a linkage matrix into a tree structure.

    Parameters
    ----------
    lm : (N, 4) array_like
        Linkage matrix.
    taxa : list of str
        Taxon names.
    rooted : bool, optional
        If True (default), will generate a rooted binary tree, with two children
        attached to the root node. If False, will generate an unrooted tree with
        three children attached to the root node.
    neg_as_zero : bool, optional
        Convert negative branch lengths to zero.

    Returns
    -------
    TreeNode
        Converted phylogenetic tree.

    See Also
    --------
    TreeNode.from_linkage_matrix

    Notes
    -----
    The linkage matrix data structure resembles that used in SciPy's hierarchical
    clustering, but also stores the branch lengths of both children, which could be
    unequal in a phylogenetic tree.

    In SciPy's linkage matrix, the four elements per row are (see:
    https://stackoverflow.com/questions/9838861/):

        cluster1, cluster2, length1+2, # original taxa

    In the current data structure, they are:

        cluster1, cluster2, length1, length2

    """
    # allocate node list
    nodes = [TreeNode(name=x) for x in taxa] + [TreeNode() for _ in range(len(lm))]

    # build tree incrementally
    idx = len(taxa)
    for c1, c2, L1, L2 in lm if rooted else lm[:-2]:
        c1, c2 = nodes[int(c1)], nodes[int(c2)]
        if neg_as_zero:
            c1.length = L1 if L1 >= 0 else 0.0
            c2.length = L2 if L2 >= 0 else 0.0
        else:
            c1.length, c2.length = L1, L2
        nodes[idx].extend([c1, c2], uncache=False)
        idx += 1

    # final treatment of an unroot tree (root is trifurcating)
    if not rooted:
        # this code assumes that the first element of the last row is the node
        # see the end of _nj
        (c0, _, L0, _), (c1, c2, L1, L2) = lm[-1], lm[-2]
        c0, c1, c2 = nodes[int(c0)], nodes[int(c1)], nodes[int(c2)]
        if neg_as_zero:
            c0.length = L0 if L0 >= 0 else 0.0
            c1.length = L1 if L1 >= 0 else 0.0
            c2.length = L2 if L2 >= 0 else 0.0
        else:
            c0.length, c1.length, c2.length = L0, L1, L2
        nodes[-1].extend([c0, c1, c2], uncache=False)
    return nodes[-1]



# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

import io
from unittest import TestCase, main

import numpy.testing as npt

from skbio import DistanceMatrix, TreeNode
from skbio.tree._nj import nj, _tree_from_linkmat


class NjTests(TestCase):

    def setUp(self):
        data1 = [[0,  5,  9,  9,  8],
                 [5,  0, 10, 10,  9],
                 [9, 10,  0,  8,  7],
                 [9, 10,  8,  0,  3],
                 [8,  9,  7,  3,  0]]
        ids1 = list('abcde')
        self.dm1 = DistanceMatrix(data1, ids1)
        # this newick string was confirmed against http://www.trex.uqam.ca/
        # which generated the following (isomorphic) newick string:
        # (d:2.0000,e:1.0000,(c:4.0000,(a:2.0000,b:3.0000):3.0000):2.0000);
        self.expected1_str = ("(d:2.000000, (c:4.000000, (b:3.000000,"
                              " a:2.000000):3.000000):2.000000, e:1.000000);")
        self.expected1_TreeNode = TreeNode.read(
                io.StringIO(self.expected1_str))

        # this example was pulled from the Phylip manual
        # https://phylipweb.github.io/phylip/doc/neighbor.html
        data2 = [[0.0000, 1.6866, 1.7198, 1.6606, 1.5243, 1.6043, 1.5905],
                 [1.6866, 0.0000, 1.5232, 1.4841, 1.4465, 1.4389, 1.4629],
                 [1.7198, 1.5232, 0.0000, 0.7115, 0.5958, 0.6179, 0.5583],
                 [1.6606, 1.4841, 0.7115, 0.0000, 0.4631, 0.5061, 0.4710],
                 [1.5243, 1.4465, 0.5958, 0.4631, 0.0000, 0.3484, 0.3083],
                 [1.6043, 1.4389, 0.6179, 0.5061, 0.3484, 0.0000, 0.2692],
                 [1.5905, 1.4629, 0.5583, 0.4710, 0.3083, 0.2692, 0.0000]]
        ids2 = ["Bovine", "Mouse", "Gibbon", "Orang", "Gorilla", "Chimp",
                "Human"]
        self.dm2 = DistanceMatrix(data2, ids2)
        self.expected2_str = ("(Mouse:0.76891, (Gibbon:0.35793, (Orang:0.28469"
                              ", (Gorilla:0.15393, (Chimp:0.15167, Human:0.117"
                              "53):0.03982):0.02696):0.04648):0.42027, Bovine:"
                              "0.91769);")
        self.expected2_TreeNode = TreeNode.read(
                io.StringIO(self.expected2_str))

        data3 = [[0, 5, 4, 7, 6, 8],
                 [5, 0, 7, 10, 9, 11],
                 [4, 7, 0, 7, 6, 8],
                 [7, 10, 7, 0, 5, 8],
                 [6, 9, 6, 5, 0, 8],
                 [8, 11, 8, 8, 8, 0]]
        ids3 = map(str, range(6))
        self.dm3 = DistanceMatrix(data3, ids3)
        self.expected3_str = ("((((0:1.000000,1:4.000000):1.000000,2:2.000000"
                              "):1.250000,5:4.750000):0.750000,3:2.750000,4:2."
                              "250000);")
        self.expected3_TreeNode = TreeNode.read(
                io.StringIO(self.expected3_str))

        # this dm can yield negative branch lengths for nj
        data4 = [[0,  5,  9,  9,  800],
                 [5,  0, 10, 10,  9],
                 [9, 10,  0,  8,  7],
                 [9, 10,  8,  0,  3],
                 [800,  9,  7,  3,  0]]
        ids4 = list('abcde')
        self.dm4 = DistanceMatrix(data4, ids4)

    def test_nj_dm1(self):
        actual_TreeNode = nj(self.dm1)
        self.assertAlmostEqual(actual_TreeNode.compare_cophenet(
            self.expected1_TreeNode), 0.0)

    def test_nj_dm2(self):
        actual_TreeNode = nj(self.dm2)
        self.assertAlmostEqual(actual_TreeNode.compare_cophenet(
            self.expected2_TreeNode), 0.0)

    def test_nj_dm3(self):
        actual_TreeNode = nj(self.dm3)
        self.assertAlmostEqual(actual_TreeNode.compare_cophenet(
            self.expected3_TreeNode), 0.0)

    def test_nj_inplace(self):
        dm = self.dm3.copy()
        obs = nj(dm)
        exp = self.expected3_TreeNode
        self.assertAlmostEqual(obs.compare_cophenet(exp), 0.0)
        npt.assert_almost_equal(dm.data, self.dm3.data)

        obs = nj(dm, inplace=True)
        self.assertAlmostEqual(obs.compare_cophenet(exp), 0.0)
        with self.assertRaises(AssertionError):
            npt.assert_almost_equal(dm.data, self.dm3.data)

    def test_nj_zero_branch_length(self):
        # no nodes have negative branch length when we disallow negative
        # branch length. self is excluded as branch length is None
        tree = nj(self.dm4)
        for n in tree.postorder(include_self=False):
            self.assertTrue(n.length >= 0)
        # only tips associated with the large distance in the input
        # have positive branch lengths when we allow negative branch
        # length
        tree = nj(self.dm4, neg_as_zero=False)
        self.assertTrue(tree.find('a').length > 0)
        self.assertTrue(tree.find('b').length < 0)
        self.assertTrue(tree.find('c').length < 0)
        self.assertTrue(tree.find('d').length < 0)
        self.assertTrue(tree.find('e').length > 0)

        # deprecated functionality
        t2 = nj(self.dm4, disallow_negative_branch_length=False)
        self.assertAlmostEqual(tree.compare_cophenet(t2), 0.0)

    def test_nj_trivial(self):
        data = [[0, 3, 2],
                [3, 0, 3],
                [2, 3, 0]]
        dm = DistanceMatrix(data, list('abc'))
        exp = TreeNode.read(["(b:2.000000, a:1.000000, c:1.000000);"])
        self.assertAlmostEqual(nj(dm).compare_cophenet(exp), 0.0)

        # deprecated functionality
        self.assertEqual(nj(dm, result_constructor=str), "(a:1.0,b:2.0,c:1.0);\n")

    def test_nj_error(self):
        data = [[0, 3],
                [3, 0]]
        dm = DistanceMatrix(data, list('ab'))
        self.assertRaises(ValueError, nj, dm)

    def test_tree_from_linkmat(self):
        taxa = ['human', 'chimp', 'monkey', 'pig', 'mouse', 'rat', 'chicken']
        lm = [
            [5, 4, 0.065041, 0.060976],   # rat, mouse
            [6, 7, 0.457317, 0.229675],   # chicken, rat-mouse
            [3, 8, 0.161924, 0.050474],   # pig, chicken-rat-mouse
            [2, 9, 0.018293, 0.162602],   # monkey, pig-...-mouse
            [10, 1, 0.006098, 0.001016],  # chimp, monkey-...-mouse
            [0, 11, 0.003049, 0.0],       # human, chimp-...-mouse
        ]
        obs = str(_tree_from_linkmat(lm, taxa, rooted=True))
        exp = ("(human:0.003049,((monkey:0.018293,(pig:0.161924,(chicken:0.457317,"
               "(rat:0.065041,mouse:0.060976):0.229675):0.050474):0.162602):0.006098,"
               "chimp:0.001016):0.0);\n")
        self.assertEqual(str(obs), exp)
        obs = str(_tree_from_linkmat(lm, taxa, rooted=False))
        exp = ("(human:0.003049,(monkey:0.018293,(pig:0.161924,(chicken:0.457317,"
               "(rat:0.065041,mouse:0.060976):0.229675):0.050474):0.162602):0.006098,"
               "chimp:0.001016);\n")
        self.assertEqual(str(obs), exp)


if __name__ == "__main__":
    main()



# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

import sys
import unittest
import inspect
import warnings

from skbio.util import classproperty
from skbio.util._decorator import (
    overrides,
    classproperty,
    classonlymethod,
    deprecated,
    aliased,
    register_aliases,
    params_aliased,
)
from skbio.util._exception import OverrideError


class TestClassOnlyMethod(unittest.TestCase):
    def test_works_on_class(self):
        class A:
            @classonlymethod
            def example(cls):
                return cls

        self.assertEqual(A.example(), A)

    def test_fails_on_instance(self):
        class A:
            @classonlymethod
            def example(cls):
                pass

        with self.assertRaises(TypeError) as e:
            A().example()

        self.assertIn('A.example', str(e.exception))
        self.assertIn('instance', str(e.exception))

    def test_matches_classmethod(self):
        class A:
            pass

        def example(cls, thing):
            """doc"""

        A.example1 = classmethod(example)
        A.example2 = classonlymethod(example)

        self.assertEqual(A.__dict__['example1'].__func__, example)
        self.assertEqual(A.__dict__['example2'].__func__, example)

        self.assertEqual(A.example1.__doc__, example.__doc__)
        self.assertEqual(A.example2.__doc__, example.__doc__)

        self.assertEqual(A.example1.__name__, example.__name__)
        self.assertEqual(A.example2.__name__, example.__name__)

    def test_passes_args_kwargs(self):
        self.ran_test = False

        class A:
            @classonlymethod
            def example(cls, arg1, arg2, kwarg1=None, kwarg2=None,
                        default=5):
                self.assertEqual(arg1, 1)
                self.assertEqual(arg2, 2)
                self.assertEqual(kwarg1, '1')
                self.assertEqual(kwarg2, '2')
                self.assertEqual(default, 5)
                self.ran_test = True

        A.example(1, *[2], kwarg2='2', **{'kwarg1': '1'})
        self.assertTrue(self.ran_test)


class TestOverrides(unittest.TestCase):
    def test_raises_when_missing(self):
        class A:
            pass

        with self.assertRaises(OverrideError):
            class B(A):
                @overrides(A)
                def test(self):
                    pass

    def test_doc_inherited(self):
        class A:
            def test(self):
                """Docstring"""
                pass

        class B(A):
            @overrides(A)
            def test(self):
                pass

        self.assertEqual(B.test.__doc__, "Docstring")

    def test_doc_not_inherited(self):
        class A:
            def test(self):
                """Docstring"""
                pass

        class B(A):
            @overrides(A)
            def test(self):
                """Different"""
                pass

        self.assertEqual(B.test.__doc__, "Different")


class TestClassProperty(unittest.TestCase):
    def test_getter_only(self):
        class Foo:
            _foo = 42

            @classproperty
            def foo(cls):
                return cls._foo

        # class-level getter
        self.assertEqual(Foo.foo, 42)

        # instance-level getter
        f = Foo()
        self.assertEqual(f.foo, 42)

        with self.assertRaises(AttributeError):
            f.foo = 4242


class TestDeprecated(unittest.TestCase):
    def test_deprecated(self):

        @deprecated("1.0", msg="It will be removed soon.")
        def foo(param):
            """I am a function.

            Parameters
            ----------
            param : str
                I am a parameter.

            """
            pass

        msg = ".. deprecated:: 1.0 It will be removed soon."
        self.assertIn(msg, foo.__doc__)
        with self.assertWarns(DeprecationWarning) as ctx:
            foo("bar")
        msg = "`foo` has been deprecated since 1.0. It will be removed soon."
        self.assertEqual(str(ctx.warning), msg)


class TestAliases(unittest.TestCase):
    def test_aliased(self):

        # apply to a function
        @aliased("bar")
        def foo(param):
            return param

        self.assertEqual(foo._alias[0], "bar")
        self.assertTrue(callable(foo._alias[1]))
        self.assertTrue(foo._alias[1]._skipdoc)
        self.assertEqual(foo.__doc__, f"Alias: ``bar``\n")
        self.assertEqual(foo(42), 42)
        self.assertEqual(foo._alias[1](42), 42)

        # apply to a class method
        class Foo:

            @aliased("bar")
            def foo(cls, param):
                return param

        self.assertEqual(Foo.foo._alias[0], "bar")

    def test_register_aliases(self):

        @register_aliases
        class Foo:

            @aliased("bar", ver="1.0", warn=True)
            def foo(self, param):
                return param

        # class level
        self.assertTrue(hasattr(Foo, "bar"))
        self.assertTrue(Foo.bar._skipdoc)

        # instance level
        f = Foo()
        self.assertTrue(hasattr(f, "bar"))

        # check warning
        self.assertFalse(hasattr(f.foo, "_warned"))
        msg = "`bar` was renamed to `foo` in 1.0."
        with self.assertWarnsRegex(DeprecationWarning, msg):
            f.bar(42)
        self.assertTrue(hasattr(f.foo, "_warned"))

    def test_params_aliased(self):

        @params_aliased([
            ("param1", "alias1", None, False),
            ("param2", "alias2", "1.0", True),
        ])
        def foo(param1, param2):
            """I am a function.

            Parameters
            ----------
            param1 : int
                The first parameter.
            param2 : int
                The second parameter.
            
            Returns
            -------
            int
                Sum of the two parameters.

            """
            return param1 + param2

        self.assertIn("Renamed from ``alias2``.", foo.__doc__)
        self.assertEqual(foo(1, 2), 3)
        self.assertEqual(foo(param1=1, param2=2), 3)
        self.assertEqual(foo(alias1=1, param2=2), 3)
        msg = "`foo`'s parameter `alias2` was renamed to `param2` in 1.0."
        with self.assertWarnsRegex(DeprecationWarning, msg):
            obs = foo(param1=1, alias2=2)
        self.assertEqual(obs, 3)
        with self.assertRaises(TypeError):
            foo(1, 2, alias1=2)


if __name__ == '__main__':
    unittest.main()



# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

from itertools import combinations

import numpy as np
import scipy.spatial.distance as spdist
from scipy.spatial.distance import squareform

from skbio.stats.distance import DistanceMatrix
from skbio.util import get_rng


def unitcorr(a, b):
    """Calculate unit correlation distance.

    d = (1 - r) / 2, in which r is the Pearson's correlation coefficient.

    """
    return spdist.correlation(a, b) * 0.5


def _check_input(trees, ids):
    """Validate input trees and IDs."""
    if not (n := len(trees)):
        raise ValueError(f"No tree is provided.")
    if ids is None:
        return
    if (m := len(ids)) != n:
        raise ValueError(f"Numbers of IDs and trees don't match.")
    if m > len(set(ids)):
        raise ValueError(f"IDs contain duplicates.")


def _check_dist_metric(metric):
    """Validate distance metric."""
    if isinstance(metric, str):
        if metric == "unitcorr":
            metric = unitcorr
        else:
            metric = getattr(spdist, metric)
    elif not callable(metric):
        raise ValueError("`metric` must be a string or callable.")
    return metric


def _check_shuffler(shuffler):
    """Validate sample shuffler."""
    if not callable(shuffler):
        shuffler = get_rng(shuffler).shuffle
    return shuffler


def _calc_dists(trees, ids, func, kwargs, shared_by_all):
    """Calculate paiwise distances among trees."""
    if shared_by_all:
        result = func(trees, **kwargs)
        result = squareform(np.array(result, dtype=float), checks=False)
        return DistanceMatrix(result, ids, validate=False)
    else:
        return DistanceMatrix.from_iterable(
            trees,
            metric=lambda x, y: func((x, y), **kwargs)[0],
            keys=ids,
            validate=False,
        )


def _setdiff(a, b, proportion):
    r"""Quantify the difference between two sets.

    For Robinson-Foulds (RF) distance calculation.

    """
    # calculate symmetric difference
    result = len(a ^ b)

    # normalize result to unit range [0, 1]
    # if total is 0, return 1 (dist = 1 means saturation)
    if proportion:
        total = len(a) + len(b)
        result = result / total if total else 1.0

    # cast result to float
    else:
        result = float(result)

    return result


def _mapdiff(a, b, metric):
    r"""Quantify the difference between two mappings.

    For weighted Robinson-Foulds (wRF) distance calculation.

    """
    # unique keys are assigned value 0
    union = frozenset(a).union(b)
    La = [a.get(x, 0.0) for x in union]
    Lb = [b.get(x, 0.0) for x in union]
    return metric(La, Lb)


def _topo_dists(
    trees,
    rooted=False,
    shared_only=True,
    proportion=False,
    include_tips=False,
    weighted=False,
    metric=None,
):
    r"""Calculate the topological distances among multiple trees.

    This function calculates the Robinson-Foulds (RF) distance or its variants. These
    metrics rely on the compatibility of branches between two trees.

    See :meth:`TreeNode.compare_rfd` and :meth:`~TreeNode.compare_wrfd` for details of
    the metrics and parameters.

    Parameters
    ----------
    trees : list of TreeNode
        The trees to compare.
    rooted : bool, optional
        Rooted or unrooted.
    shared_only : bool, optional
        Refine to shared taxa.
    proportion : bool, optional
        Normalize to proportion.
    include_tips : bool, optional
        Include singletons.
    weighted : bool, optional
        Weight by branch length.
    metric : callable, optional
        Distance metric (must provide if weighted).

    Returns
    -------
    list of float
        Pairwise path-length distances or variants in condensed form.

    See Also
    --------
    rf_dists
    wrfd_dists
    TreeNode.compare_rfd
    TreeNode.compare_wrfd
    TreeNode.compare_subsets
    TreeNode.compare_biparts

    """
    kwargs = dict(include_tips=include_tips, map_to_length=weighted)

    # get topological units reflecting shared taxa
    if shared_only:
        taxon_sets = [t.subset() for t in trees]
        shared = frozenset.intersection(*taxon_sets)
        n_shared = len(shared)

        # Decide the "within" parameter (None or shared taxa) for each method call,
        # which is more efficient when within=None.
        withins = [shared if len(s) > n_shared else None for s in taxon_sets]

        # get topological units from all trees
        if rooted:
            sets = [t.subsets(within=w, **kwargs) for t, w in zip(trees, withins)]
        else:
            sets = [
                t.biparts(within=w, full=s, **kwargs)
                for t, w, s in zip(trees, withins, taxon_sets)
            ]

    # get all topological units
    else:
        sets = [(t.subsets if rooted else t.biparts)(**kwargs) for t in trees]

    # unweighted (symmetric difference between sets)
    if not weighted:
        result = [_setdiff(a, b, proportion) for a, b in combinations(sets, 2)]

    # branch length weighted (vector distance)
    else:
        result = [_mapdiff(a, b, metric) for a, b in combinations(sets, 2)]

    return result


def rf_dists(trees, ids=None, shared_by_all=True, proportion=False, rooted=False):
    r"""Calculate Robinson-Foulds (RF) distances among trees.

    .. versionadded:: 0.6.3

    Parameters
    ----------
    trees : list of TreeNode
        Input trees.
    ids : list of str, optional
        Unique identifiers of input trees. If omitted, will use incremental integers
        "0", "1", "2",...
    shared_by_all : bool, optional
        Calculate the distance between each pair of trees based on taxa shared across
        all trees (True, default), or shared between the current pair of trees (False).
    proportion : bool, optional
        Whether to return the RF distance as count (False, default) or proportion
        (True).
    rooted : bool, optional
        Whether to consider the trees as unrooted (False, default) or rooted (True).

    Returns
    -------
    DistanceMatrix
        Matrix of the Robinson-Foulds distances.

    See Also
    --------
    TreeNode.compare_rfd

    Notes
    -----
    The Robinson-Foulds (RF) distance [1]_, a.k.a. symmetric difference, is the number
    of bipartitions differing between two trees.

    This function is equivalent to :meth:`TreeNode.compare_rfd` for two trees. Refer
    to the latter for details about the metric and its parameters. However, the current
    function extends the operation to an arbitrary number of trees and returns a
    distance matrix for them.

    This function is optimized for calculation based on taxa shared across all trees.
    One can instead set ``shared_by_all`` to False to calculate based on taxa shared
    between each pair of trees, which is however less efficient since bipartitions need
    to be re-inferred during each comparison.

    References
    ----------
    .. [1] Robinson, D. F., & Foulds, L. R. (1981). Comparison of phylogenetic trees.
       Mathematical biosciences, 53(1-2), 131-147.

    Examples
    --------
    >>> from skbio import TreeNode
    >>> trees = [TreeNode.read([x]) for x in (
    ...     "(((a,b),c),d,e);",
    ...     "((a,(b,c)),d,e);",
    ...     "((a,b),(c,d),e);",
    ...     "(a,b,(c,(d,e)));",
    ... )]
    >>> dm = rf_dists(trees, ids=list("ABCD"))
    >>> print(dm)
    4x4 distance matrix
    IDs:
    'A', 'B', 'C', 'D'
    Data:
    [[ 0.  2.  2.  0.]
     [ 2.  0.  4.  2.]
     [ 2.  4.  0.  2.]
     [ 0.  2.  2.  0.]]

    """
    _check_input(trees, ids)
    kwargs = dict(
        rooted=rooted,
        shared_only=True,
        proportion=proportion,
        include_tips=False,
        weighted=False,
        metric=None,
    )
    return _calc_dists(trees, ids, _topo_dists, kwargs, shared_by_all)


def wrf_dists(
    trees,
    ids=None,
    shared_by_all=True,
    metric="cityblock",
    rooted=False,
    include_tips=True,
):
    r"""Calculate weighted Robinson-Foulds (wRF) distances or variants among trees.

    .. versionadded:: 0.6.3

    Parameters
    ----------
    trees : list of TreeNode
        Input trees.
    ids : list of str, optional
        Unique identifiers of input trees. If omitted, will use incremental integers
        "0", "1", "2",...
    shared_by_all : bool, optional
        Calculate the distance between each pair of trees based on taxa shared across
        all trees (True, default), or shared between the current pair of trees (False).
    metric : str or callable, optional
        The distance metric to use. Can be a preset, a distance function name under
        :mod:`scipy.spatial.distance`, or a custom function that takes two vectors and
        returns a number. See :meth:`~TreeNode.compare_wrfd` for details.
    rooted : bool, optional
        Whether to consider the trees as unrooted (False, default) or rooted (True).
    include_tips : bool, optional
        Whether to include single-taxon biparitions (terminal branches) in the
        calculation. Default is True.

    Returns
    -------
    DistanceMatrix
        Matrix of weighted Robinson-Foulds distances or variants.

    See Also
    --------
    TreeNode.compare_wrfd

    Notes
    -----
    The weighted Robinson-Foulds (wRF) distance [1]_ is the sum of differences of
    branch lengths of matching bipartitions between a pair of trees.

    This function is equivalent to :meth:`TreeNode.compare_wrfd` for two trees. Refer
    to the latter for details about the metric and its variants, and the parameter
    settings for calculating them. However, the current function extends the operation
    to an arbitrary number of trees and returns a distance matrix for them.

    A restriction of the current function compared to ``compare_wrfd`` is that
    ``metric`` must be symmetric (i.e., :math:`d(x, y) = d(y, x)`), and equals zero
    from a vector to itself (i.e., :math:`d(x, x) = 0`). It does not have to suffice
    non-negativity or triangle inequality though.

    This function is optimized for calculation based on taxa shared across all trees.
    One can instead set ``shared_by_all`` to False to calculate based on taxa shared
    between each pair of trees, which is however less efficient as bipartitions need to
    be re-inferred during each comparison.

    References
    ----------
    .. [1] Robinson, D. F., & Foulds, L. R. (1979) Comparison of weighted labelled
       trees. In Combinatorial Mathematics VI: Proceedings of the Sixth Australian
       Conference on Combinatorial Mathematics, Armidale, Australia (pp. 119-126).

    Examples
    --------
    >>> from skbio import TreeNode
    >>> trees = [TreeNode.read([x]) for x in (
    ...     "((a:1,b:2):1,c:4,((d:4,e:5):2,f:6):1);",
    ...     "((a:3,(b:2,c:2):1):3,d:8,(e:5,f:6):2);",
    ...     "((a:1,c:6):2,(b:3,(d:2,e:3):1):2,f:7);",
    ... )]
    >>> dm = wrf_dists(trees, ids=list("ABC"))
    >>> print(dm)
    3x3 distance matrix
    IDs:
    'A', 'B', 'C'
    Data:
    [[  0.  16.  15.]
     [ 16.   0.  27.]
     [ 15.  27.   0.]]

    """
    _check_input(trees, ids)
    metric = _check_dist_metric(metric)
    kwargs = dict(
        rooted=rooted,
        shared_only=True,
        include_tips=include_tips,
        weighted=True,
        metric=metric,
    )
    return _calc_dists(trees, ids, _topo_dists, kwargs, shared_by_all)


def _path_dists(trees, sample, metric, shuffler, use_length, ignore_self):
    r"""Calculate path-length distances or variants among multiple trees.

    This function calculates several tree distance metrics based on the lengths of
    paths connecting pairs of taxa, a.k.a., cophenetic distances.

    See :meth:`TreeNode.compare_cophenet` for details of the metrics and parameters.

    Parameters
    ----------
    trees : list of TreeNode
        The trees to compare.
    sample : int, optional
        Number of taxa to sample.
    metric : callable, optional
        Distance function.
    shuffler : callable, optional
        Shuffling function.
    use_length : bool, optional
        Use branch lengths.
    ignore_self : bool, optional
        Ignore taxon to itself.

    Returns
    -------
    list of float
        Pairwise path-length distances or variants in condensed form.

    See Also
    --------
    path_dists
    TreeNode.compare_cophenet

    """
    # Get shared taxa among trees.
    shared = frozenset.intersection(*(t.subset() for t in trees))

    # Sort shared taxa. This is for the stability of sampling and distance calculation
    # as Python set is unordered.
    shared = sorted(shared)

    # Randomly sample a subset of taxa.
    if sample is not None:
        if (n_shared := len(shared)) < sample:
            raise ValueError(
                f"Sample size ({sample}) is greater than the number of shared taxa "
                f"({n_shared})."
            )
        shuffler(shared)
        shared = shared[:sample]

    # Generate cophenetic (tip-to-tip) distance matrices.
    paths = [x.cophenet(endpoints=shared, use_length=use_length) for x in trees]

    # Convert square matrices into condensed 1-D arrays, such that diagonal (d(x, x) =
    # 0) and symmetric (d(x, y) = d(y, x)) comparisons are omitted.
    if ignore_self:
        paths = [x.condensed_form() for x in paths]

    # Otherwise, create 1-D views of the entire square matrics. `ravel` copies the data
    # only when necessary, and in this case doesn't.
    else:
        paths = [x.data.ravel() for x in paths]

    # Calculate pairwise distances among matrices.
    return [metric(a, b) for a, b in combinations(paths, 2)]


def path_dists(
    trees,
    ids=None,
    shared_by_all=True,
    metric="euclidean",
    use_length=True,
    sample=None,
    shuffler=None,
):
    r"""Calculate path-length distances or variants among trees.

    .. versionadded:: 0.6.3

    Parameters
    ----------
    trees : list of TreeNode
        Input trees.
    ids : list of str, optional
        Unique identifiers of input trees. If omitted, will use incremental integers
        "0", "1", "2",...
    shared_by_all : bool, optional
        Calculate the distance between each pair of trees based on taxa shared across
        all trees (True, default), or shared between the current pair of trees (False).
    metric : str or callable, optional
        The distance metric to use. Can be a preset, a distance function name under
        :mod:`scipy.spatial.distance`, or a custom function that takes two vectors and
        returns a number. See :meth:`~TreeNode.compare_cophenet` for details.
    use_length : bool, optional
        Whether to calculate the sum of branch lengths (True, default) or the
        number of branches (False) connecting each pair of tips.
    sample : int, optional
        Randomly subsample this number of shared taxa for calculation. This is useful
        when comparing very large trees.
    shuffler : int, np.random.Generator or callable, optional
        The shuffling function to use if ``sample`` is specified. Default is
        :meth:`~numpy.random.Generator.shuffle`. If an integer is provided, a random
        generator will be constructed using this number as the seed.

    Returns
    -------
    DistanceMatrix
        Matrix of the path-length distances or variants.

    See Also
    --------
    TreeNode.compare_cophenet

    Notes
    -----
    The path-length distance [1]_ is the square root of the sum of squared differences
    of path lengths among all pairs of taxa between two trees.

    This function is equivalent to :meth:`TreeNode.compare_cophenet` for two trees.
    Refer to the latter for details about the metric and its variants, and parameter
    settings for calculating them. However, the current function extends the operation
    to an arbitrary number of trees and returns a distance matrix for them. It is named
    so because the term "cophenetic distance" refers to the distance between two taxa
    in a tree instead.

    A restriction of the current function compared to ``compare_cophenet`` is that
    ``metric`` must be symmetric (i.e., :math:`d(x, y) = d(y, x)`), and equals zero
    from a vector to itself (i.e., :math:`d(x, x) = 0`). It does not have to suffice
    non-negativity or triangle inequality though.

    This function is optimized for calculation based on taxa shared across all trees.
    One can instead set ``shared_by_all`` to False to calculate based on taxa shared
    between each pair of trees, which is however less efficient as the path lengths
    need to be re-calculated during each comparison.

    References
    ----------
    .. [1] Lapointe, F. J., & Cucumel, G. (1997). The average consensus procedure:
       combination of weighted trees containing identical or overlapping sets of
       taxa. Systematic Biology, 46(2), 306-312.

    Examples
    --------
    >>> from skbio import TreeNode
    >>> trees = [TreeNode.read([x]) for x in (
    ...     "((a:1,b:2):1,c:4,((d:4,e:5):2,f:6):1);",
    ...     "((a:3,(b:2,c:2):1):3,d:8,(e:5,f:6):2);",
    ...     "((a:1,c:6):2,(b:3,(d:2,e:3):1):2,f:7);",
    ... )]
    >>> dm = path_dists(trees, ids=list("ABC"))
    >>> print(dm)
    3x3 distance matrix
    IDs:
    'A', 'B', 'C'
    Data:
    [[  0.          13.7113092   11.87434209]
     [ 13.7113092    0.          19.5192213 ]
     [ 11.87434209  19.5192213    0.        ]]

    """
    _check_input(trees, ids)
    metric = _check_dist_metric(metric)
    if sample is not None:
        shuffler = _check_shuffler(shuffler)
    kwargs = dict(
        sample=sample,
        metric=metric,
        shuffler=shuffler,
        use_length=use_length,
        ignore_self=True,
    )
    return _calc_dists(trees, ids, _path_dists, kwargs, shared_by_all)



# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

from scipy.cluster.hierarchy import linkage
from skbio.tree import TreeNode
from ._utils import _check_dm


def upgma(dm, weighted=False):
    r"""Perform unweighted pair group method with arithmetic mean (UPGMA) or its
    weighted variant (WPGMA) for phylogenetic reconstruction.

    Parameters
    ----------
    dm : skbio.DistanceMatrix
        The input distance matrix.
    weighted : bool, optional
        If True, WPGMA is performed instead of UPGMA. WPGMA is a variant of UPGMA
        which is unbiased towards the size of subtrees computed.

    Returns
    -------
    TreeNode
        A TreeNode object with estimated edge values.

    See Also
    --------
    nj

    Notes
    -----
    UPGMA (unweighted pair group method with arithmetic mean) is a simple hierarchical
    clustering method that iteratively groups proximal taxa or taxon groups to form a
    tree structure. A weighted variant is known as WPGMA, and both variants are due to
    Sokal and Michener [1]_.

    This function wraps SciPy's :func:`~scipy.cluster.hierarchy.linkage` function, with
    the ``method`` parameter set as "average" (UPGMA) or "weighted" (WPGMA). It takes a
    scikit-bio DistanceMatrix object and returns a scikit-bio TreeNode object.

    UPGMA creates a rooted and ultrametric tree -- all tips will have the same height
    (distance from the root node).

    References
    ----------
    .. [1] Sokal, R.R., & Michener, C.D. (1958). A statistical method for
       evaluating systematic relationships. University of Kansas science
       bulletin, 38, 1409-1438.

    Examples
    --------
    Define a distance matrix object for the taxa a, b, and c.

    >>> from skbio import DistanceMatrix

    >>> data = [[0, 1, 2],
    ...         [1, 0, 3],
    ...         [2, 3, 0]]
    >>> ids = list('abc')
    >>> dm = DistanceMatrix(data, ids)

    Construct a tree using UPGMA.

    >>> tree = upgma(dm)
    >>> print(tree.ascii_art())
              /-c
    ---------|
             |          /-a
              \--------|
                        \-b

    The tree also has estimated edge values assigned to each edge.

    >>> print(tree)
    (c:1.25,(a:0.5,b:0.5):0.75);
    <BLANKLINE>
    """
    _check_dm(dm)

    # If weighted is set to 'False', UPGMA is performed
    if weighted is False:
        linkage_matrix = linkage(
            dm.condensed_form(), method="average", metric="euclidean"
        )
    # Otherwise, WPGMA is performed
    else:
        linkage_matrix = linkage(
            dm.condensed_form(), method="weighted", metric="euclidean"
        )

    # Construct the TreeNode from the linkage matrix
    tree = TreeNode.from_linkage_matrix(linkage_matrix, dm.ids)

    return tree



# ----------------------------------------------------------------------------
# Copyright (c) 2013--, scikit-bio development team.
#
# Distributed under the terms of the Modified BSD License.
#
# The full license is in the file LICENSE.txt, distributed with this software.
# ----------------------------------------------------------------------------

# cython: language_level=3, boundscheck=False, wraparound=False, cdivision=True
# distutils: define_macros=NPY_NO_DEPRECATED_API=NPY_1_7_API_VERSION

cimport cython
import numpy as np
cimport numpy as cnp
cnp.import_array()
from cython.parallel import prange
from heapq import heappush


# This script hosts components of the greedy algorithms for phylogenetic reconstruction
# using the minimum evolution (ME) principle. Specifically, it supports GME and BME for
# de novo tree building, and FastNNI and BNNI for tree arrangement.

# Most functions don't create any intermediate arrays, but repeatedly use pre-allocated
# arrays for all operations. This improves computational efficiency.


# chunk size for parallelization (experimental)
cdef int CHUNKSIZE = 10000

# ------------------------------------------------------------------------------------
# NOTE on parallelization: The algorithms implemented here were designed to facilitate
# parallelization. Specifically, the preorder and postorder are stored, with which the
# code can identify the entire clade under any given node, without needing to redo the
# tree traversal. Operations that are topology-independent can be parallelized across
# nodes within the clade.
# 
# A challenge to parallelization is scheduling. Clades vary greatly in size, therefore
# iterations also have greatly varying compute loads. It is tricky to allocate chunks
# of iterations to individual threads such that all threads have roughly the same load.
# In the experimental code `_bal_avgdist_insert_p`, a "dynamic" policy with a chunk
# size inversely proportional to the clade size is adopted, thus fewer larger clades
# will be processed in each thread.
#
# Additionally, it would have been ideal to disable parallelization if the clade being
# processed is small. In a typical traversal, half of the clades are tips, making it
# useless to parallelize. Cython 3.1 will introduce a parameter `use_threads_if`, which
# can address this issue. But we need to put it on hold before Cython 3.1 is released
# and widely adopted.
# ------------------------------------------------------------------------------------


def _preorder(
    Py_ssize_t[::1] order,
    Py_ssize_t[:, ::1] tree,
    Py_ssize_t[::1] stack,
    Py_ssize_t start = 0,
):
    r"""Perform preorder traversal.

    This function and :func:`_postorder` use stacks to avoid recursion. The stack array
    is pre-allocated. The output (ordered nodes) is also written into a pre-allocated
    array.

    This function and :func:`_postorder` are not actually used in the greedy algorithms
    in this module, which incrementally grow the tree as well as the orders. The two
    functions are implemented for reference and test purpose.

    """
    cdef Py_ssize_t curr, left
    cdef Py_ssize_t order_i = 0  # next index of order
    cdef Py_ssize_t stack_i = 1  # next index of stack
    stack[0] = start
    while stack_i:
        # pop a node from stack into order
        stack_i -= 1
        curr = stack[stack_i]
        order[order_i] = curr
        # index[curr] = order_i  # preorder index
        order_i += 1
        # append children to stack, right first such that left is processed first
        left = tree[curr, 0]
        if left:
            stack[stack_i] = tree[curr, 1]
            stack[stack_i + 1] = left
            stack_i += 2


def _postorder(
    Py_ssize_t[::1] order,
    Py_ssize_t[:, ::1] tree,
    Py_ssize_t[::1] stack,
    Py_ssize_t start = 0,
):
    """Perform postorder traversal.

    See also :func:`_preorder`.

    """
    cdef Py_ssize_t curr, last, left, right
    cdef Py_ssize_t order_i = 0
    cdef Py_ssize_t stack_i = 1
    cdef Py_ssize_t prev = 0  # last visited node
    stack[0] = start
    curr = tree[start, 0]
    while stack_i:
        if curr:
            stack[stack_i] = curr
            stack_i += 1
            curr = tree[curr, 0]  # go to left child
        else:
            last = stack[stack_i - 1]
            left, right = tree[last, 0], tree[last, 1]
            if left and prev != right:
                curr = right  # go to right child
            else:
                order[order_i] = last
                # index[last] = order_i  # postorder index
                order_i += 1
                prev = last
                stack_i -= 1


def _avgdist_matrix(
    double[:, ::1] adm,
    double[:, :] dm,
    Py_ssize_t[:, ::1] tree,
    Py_ssize_t[::1] preodr,
    Py_ssize_t[::1] postodr,
):
    r"""Calculate a matrix of average distances between all pairs of subtrees.

    This function will update adm, a float array of (n, n) representing pairwise
    distances between all nodes (tips, internal nodes and the root) in the tree.

    Implemented according to Appendix 4 of Desper and Gascuel (2002). Basically, this
    algorithm traverses the tree and calculates the average distance between each pair
    of subtrees. Here, a "subtree" is identified by a node, but it can be one of the
    two scenarios:

    0. Lower subtree (L): the subtree descending from a node (including the node).
    1. Upper subtree (U): the subtree branching from the node upward. The root of this
       subtree is the node's parent (NOT the node itself). Its immediate children are
       the parent's parent and the node's sibling.

    Then it iteratively applies Eq. 2 to calculate the average distance between two
    subtrees based on the average distances from the children (1 and 2) of one of the
    subtrees (B) to the other (A):

        d(A, B) = (|B_1| * d(A, B_1) + |B_2| * d(A, B_2)) / |B|

    TODO: It might be possible to only fill half of the matrix (upper or lower
    triangle). Same for other functions, especially those in BME.

    """
    cdef Py_ssize_t i, j, k
    cdef Py_ssize_t curr, parent, sibling, p_size, s_size
    cdef Py_ssize_t a, a_size, a_taxon
    cdef Py_ssize_t a1 = 0, a2 = 0, a1_size = 0, a2_size = 0
    cdef Py_ssize_t b, b_size, b1, b2

    # total numbers of taxa and nodes in the tree
    cdef Py_ssize_t m = tree[0, 4] + 1
    cdef Py_ssize_t n = 2 * m - 3

    # Calculate the average distance between each pair of subtrees defined by nodes
    # a and b (A4.1 of the paper).

    # Step 1: Calculate non-nested subtree-to-subtree distances (i.e., one is not an
    # ancestor of another. Therefore each subtree is the lower (descending) tree of
    # the node. (A4.1 (a))
    # The paper suggests outer and inner postorder traverals for all nodes. Since
    # distances are symmetric, one can save half of the calculations, as done by the
    # following code.

    # Loop over nodes in postorder.
    # Skip the root, which is always the last node in the postorder traversal.
    for i in range(n - 1):
        a = postodr[i]
        a_size = tree[a, 4]

        # check if a is a tip
        if a_size == 1:
            a_taxon = tree[a, 1]
        else:
            a_taxon = 0  # a can never be root (taxon 0), therefore 0 means none
            a1, a2 = tree[a, 0], tree[a, 1]
            a1_size, a2_size = tree[a1, 4], tree[a2, 4]

        # Iterate over the ancestors of a, and find the other subtree on the right.
        # We can skip the left subtree because it must have been calculated already.
        # If the right subtree contains a's ancestry, just skip.
        curr = a
        while curr:
            parent = tree[curr, 2]
            if tree[parent, 0] != curr:
                curr = parent
                continue
            sibling = tree[curr, 3]

            # Loop over nodes within the right subtree in postorder.
            # This postorder doesn't need to be re-calculated. Because all nodes within
            # a clade are continuous in postorder, one can take a slice of the full
            # postorder that represent the descending nodes of the current node. The
            # size of the slice is 2 x taxon count - 2. *
            k = tree[sibling, 7]
            for j in range(k - tree[sibling, 4] * 2 + 2, k + 1):
                b = postodr[j]
                b_size = tree[b, 4]

                # If both a and b are tips, take the original taxon-to-taxon distance
                # (A4.1 (a) i).
                if b_size == 1 and a_taxon != 0:
                    adm[a, b] = adm[b, a] = dm[a_taxon, tree[b, 1]]

                # If a is an internal node, and b is either (a tip or an internal node),
                # calculate the average distance based on the two child subtrees of a
                # (A4.1 (a) ii).
                elif a_taxon == 0:
                    adm[a, b] = adm[b, a] = (
                        a1_size * adm[a1, b] + a2_size * adm[a2, b]
                    ) / a_size

                # If a is a tip, and b is an internal node, calculate the average
                # distance based on the two child subtrees of b (A4.1 (a) iii).
                else:
                    b1, b2 = tree[b, 0], tree[b, 1]
                    adm[a, b] = adm[b, a] = (
                        tree[b1, 4] * adm[a, b1] + tree[b2, 4] * adm[a, b2]
                    ) / b_size

            curr = parent

    # Step 2: Calculate subtree to root (taxon 0) distances (A4.1 (b)).

    # This is done through a postorder traversal.
    for i in range(n - 1):
        a = postodr[i]
        a_size = tree[a, 4]
        if a_size == 1:
            a_taxon = tree[a, 1]
            adm[a, 0] = adm[0, a] = dm[0, a_taxon]
        else:
            a1, a2 = tree[a, 0], tree[a, 1]
            adm[a, 0] = adm[0, a] = (
                tree[a1, 4] * adm[a1, 0] + tree[a2, 4] * adm[a2, 0]
            ) / a_size

    # Step 3: Calculate nested subtree to subtree distances, in which the first node
    # (a) is a descendant of the second node (b), therefore the first subtree (A) is
    # the lower (descending) tree of node a, whereas the second subtree (B) is the
    # upper (ancestral) tree of node b (A4.1 (c)).

    # This is done through a preorder traversal.
    for i in range(1, n):
        a = preodr[i]
        parent, sibling = tree[a, 2], tree[a, 3]

        # The size of (upper) subtree b is the complement of its descendants. Same for
        # the parent subtree.
        a_size = m - tree[a, 4]
        p_size = m - tree[parent, 4]
        s_size = tree[sibling, 4]

        # Iterate over all subtrees below b.
        # The paper says this traversal can be done in any manner. Here, we use the
        # postorder. See * above.
        k = tree[a, 7]
        for j in range(k - tree[a, 4] * 2 + 2, k):
            b = postodr[j]
            adm[a, b] = adm[b, a] = (
                s_size * adm[b, sibling] + p_size * adm[b, parent]
            ) / a_size


def _bal_avgdist_matrix(
    double[:, ::1] adm,
    double[:, :] dm,
    Py_ssize_t[:, ::1] tree,
    Py_ssize_t[::1] preodr,
    Py_ssize_t[::1] postodr,
):
    r"""Calculate a matrix of balanced average distances between all pairs of subtrees.

    This function resembles :func:`_avgdist_matrix`, but it weighs subtrees equally
    regardless of their sizes. Specifically, it replaces Eq. 2 with Eq. 6. of Desper
    and Gascuel (2002):

        d(A, B) = (d(A, B_1) + d(A, B_2)) / 2

    Same for all functions starting with `bal_`.

    """
    cdef Py_ssize_t i, j, k
    cdef Py_ssize_t curr, parent, sibling
    cdef Py_ssize_t a, b, a_taxon
    cdef Py_ssize_t a1 = 0, a2 = 0

    cdef Py_ssize_t n = 2 * tree[0, 4] - 1

    # Step 1: Calculate non-nested subtree to subtree distances.
    for i in range(n - 1):
        a = postodr[i]
        if tree[a, 0] == 0:
            a_taxon = tree[a, 1]
        else:
            a_taxon = 0
            a1, a2 = tree[a, 0], tree[a, 1]
        curr = a
        while curr:
            parent = tree[curr, 2]
            if tree[parent, 0] != curr:
                curr = parent
                continue
            sibling = tree[curr, 3]
            k = tree[sibling, 7]
            for j in range(k - tree[sibling, 4] * 2 + 2, k + 1):
                b = postodr[j]
                if tree[b, 0] == 0 and a_taxon != 0:
                    adm[a, b] = adm[b, a] = dm[a_taxon, tree[b, 1]]
                elif a_taxon == 0:
                    adm[a, b] = adm[b, a] = 0.5 * (adm[a1, b] + adm[a2, b])
                else:
                    adm[a, b] = adm[b, a] = 0.5 * (
                        adm[a, tree[b, 0]] + adm[a, tree[b, 1]]
                    )
            curr = parent

    # Step 2: Calculate subtree to root distances.
    for i in range(n - 1):
        a = postodr[i]
        if tree[a, 0] == 0:
            adm[a, 0] = adm[0, a] = dm[0, tree[a, 1]]
        else:
            adm[a, 0] = adm[0, a] = 0.5 * (adm[tree[a, 0], 0] + adm[tree[a, 1], 0])

    # Step 3: Calculate nested subtree to subtree distances.
    for i in range(1, n):
        a = preodr[i]
        parent, sibling, k = tree[a, 2], tree[a, 3], tree[a, 7]
        for j in range(k - tree[a, 4] * 2 + 2, k):
            b = postodr[j]
            adm[a, b] = adm[b, a] = 0.5 * (adm[b, sibling] + adm[b, parent])


def _avgdist_taxon(
    double[:, ::1] adk,
    Py_ssize_t taxon,
    double[:, :] dm,
    Py_ssize_t[:, ::1] tree,
    Py_ssize_t[::1] preodr,
    Py_ssize_t[::1] postodr,
):
    """Calculate average distances between a new taxon (k) and existing subtrees.

    This function will update adk, a float array of (n, 2) in which columns 0 and 1
    represent the average distances from the taxon to the lower and upper subtrees of
    each existing node, respectively.

    Implemented according to Appendix 3 of Desper and Gascuel (2002). Basically, this
    algorithm calculates all lower subtree distances via a postorder traversal, then
    calculates all upper subtree distances via a preorder traversal.

    """
    cdef Py_ssize_t i
    cdef Py_ssize_t node, left, right, parent, sibling

    # total numbers of taxa and nodes in the tree
    cdef Py_ssize_t m = tree[0, 4] + 1
    cdef Py_ssize_t n = 2 * m - 3

    # Calculate the distance between taxon and the lower subtree of each node.
    for i in range(n):
        node = postodr[i]
        left, right = tree[node, 0], tree[node, 1]
        if left == 0:
            adk[node, 0] = dm[taxon, right]
        else:
            adk[node, 0] = (
                tree[left, 4] * adk[left, 0] + tree[right, 4] * adk[right, 0]
            ) / tree[node, 4]

    # Assign upper distance of root.
    adk[0, 1] = dm[taxon, 0]

    # Calculate the distance between taxon k and the upper subtree of each node.
    for i in range(1, n):
        node = preodr[i]
        parent, sibling = tree[node, 2], tree[node, 3]
        adk[node, 1] = (
            (m - tree[parent, 4]) * adk[parent, 1] + tree[sibling, 4] * adk[sibling, 0]
        ) / (m - tree[node, 4])


def _bal_avgdist_taxon(
    double[:, ::1] adk,
    Py_ssize_t taxon,
    double[:, :] dm,
    Py_ssize_t[:, ::1] tree,
    Py_ssize_t[::1] preodr,
    Py_ssize_t[::1] postodr,
):
    r"""Calculate balanced average distances between a new taxon and existing subtrees.

    This function resembles :func:`_avgdist_taxon` but uses the balanced framework.

    """
    cdef Py_ssize_t i
    cdef Py_ssize_t node, left, right
    cdef Py_ssize_t n = 2 * tree[0, 4] - 1
    for i in range(n):
        node = postodr[i]
        left, right = tree[node, 0], tree[node, 1]
        if left == 0:
            adk[node, 0] = dm[taxon, right]
        else:
            adk[node, 0] = 0.5 * (adk[left, 0] + adk[right, 0])
    adk[0, 1] = dm[taxon, 0]
    for i in range(1, n):
        node = preodr[i]
        adk[node, 1] = 0.5 * (adk[tree[node, 2], 1] + adk[tree[node, 3], 0])


def _ols_lengths(
    double[::1] lens,
    double[:, ::1] adm,
    Py_ssize_t[:, ::1] tree,
):
    r"""Calculate branch lengths of a tree based on the OLS framework.

    Using an average distance matrix between all pairs of subtrees.

    Implemented according to Eqs. 3 & 4 of Desper and Gascuel (2002).

    TODO: Can be parallelized. Although this isn't a bottlenecking step.

    """
    cdef Py_ssize_t node, left, right, parent, sibling
    cdef Py_ssize_t l_size, r_size, p_size, s_size
    cdef double lambda_
    cdef Py_ssize_t m = tree[0, 4] + 1

    for node in range(1, 2 * m - 3):
        left = tree[node, 0]
        parent = tree[node, 2]
        sibling = tree[node, 3]

        # External (terminal) branch: based on the triplet (iAB) of self (i), parent
        # (A), and sibling (B) (Eq. 4).
        if left == 0:
            lens[node] = 0.5 * (
                adm[parent, node] + adm[sibling, node] - adm[parent, sibling]
            )

        # Internal branch: based on the quartet (AB|CD) of parent (A, upper), sibling
        # (B), and children (C and D) (Eq. 3).
        else:
            right = tree[node, 1]
            l_size = tree[left, 4]
            r_size = tree[right, 4]
            p_size = m - tree[parent, 4]
            s_size = tree[sibling, 4]
            lambda_ = <double>(p_size * r_size + s_size * l_size) / (
                (p_size + s_size) * (l_size + r_size)
            )
            lens[node] = 0.5 * (
                lambda_ * (adm[parent, left] + adm[sibling, right])
                + (1 - lambda_) * (adm[parent, right] + adm[sibling, left])
                - (adm[parent, sibling] + adm[left, right])
            )

    # root branch
    left, right = tree[0, 0], tree[0, 1]
    lens[0] = 0.5 * (adm[left, 0] + adm[right, 0] - adm[left, right])


def _ols_lengths_d2(
    double[::1] lens,
    double[:, ::1] ad2,
    Py_ssize_t[:, ::1] tree,
):
    r"""Calculate branch lengths of a tree based on an OLS framework.

    Using only average distances between pairs of distant-2 subtrees.

    This function produces the same result as `_ols_lengths`. The latter relies on
    Eq. 3 of Desper and Gascuel (2002), which involves distances between distant-3
    subtrees. Therefore, we need to modify the equation such that it only takes
    distances between distant-2 subtrees as input.

    Specifically, with the following local structure:

                |
              parent
              /   \
           node  sibling
           /  \
        left  right

    We will need to calculate the distances between left / right (lower) and parent
    (upper) / sibling (lower). This can be achieved by (according to Fig. 2a):

        d(left(L), parent(U)) = d(left(L), node(U)) + d(node(L), parent(U))
            - d(node(L), node(U))

    Do the same for d(right(L), parent(U)), d(left(L), sibling(L)), and d(right(L),
    sibling(L)). Plug results into Eq. 3, we will get:

        l(node) = 0.5 * (
            d(left(L), parent(U)) + d(right(L), parent(U)) + d(node(L), parent(U))
            + d(node(L), sibling(U)) - d(sibling(L), parent(U)) - d(left(L), right(L))
        ) - d(node(L), node(U))

    Note that this equation is free of the lambda factor.

    Here, d(node(L), node(U)) is the distance between the lower and upper subtrees of
    the same node. This can be calculated using the equation in A4.1(c):

        d(node(L), node(U)) = (|sibling(L)| * d(node(L), sibling(L)) + |parent(U)|
            * d(node(L), parent(U))) / |node(U)|

    Therefore, we will get l(node).

    """
    cdef Py_ssize_t node, left, sibling
    cdef Py_ssize_t m = tree[0, 4] + 1

    for node in range(1, 2 * m - 3):
        left = tree[node, 0]

        # External (terminal) branch: based on Eq. 4 of the paper.
        if left == 0:
            lens[node] = 0.5 * (ad2[node, 0] + ad2[node, 1] - ad2[tree[node, 3], 0])

        # Internal branch: based on the equation discussed above.
        else:
            sibling = tree[node, 3]
            lens[node] = 0.5 * (
                ad2[left, 0]
                + ad2[tree[node, 1], 0]
                + ad2[node, 0]
                + ad2[node, 1]
                - ad2[sibling, 0]
                - ad2[left, 1]
            ) - (
                tree[sibling, 4]
                * ad2[node, 1]
                + (m - tree[tree[node, 2], 4])
                * ad2[node, 0]
            ) / (m - tree[node, 4])

    # root branch
    left = tree[0, 0]
    lens[0] = 0.5 * (ad2[left, 0] + ad2[tree[0, 1], 0] - ad2[left, 1])


def _bal_lengths(
    double[::1] lens,
    double[:, ::1] adm,
    Py_ssize_t[:, ::1] tree,
):
    r"""Calculate branch lengths of a tree based on the balanced framework.

    Using a balanced average distance matrix between all pairs of subtrees.

    This function resembles :func:`_ols_lengths` but it uses the balanced framework.

    Implemented according to Eqs. 3 & 4 and the description on top of pg. 691 of Desper
    and Gascuel (2002).

    """
    cdef Py_ssize_t node, left, right, parent, sibling

    for node in range(1, 2 * tree[0, 4] - 1):
        left = tree[node, 0]
        parent = tree[node, 2]
        sibling = tree[node, 3]
        if left == 0:
            lens[node] = 0.5 * (
                adm[parent, node] + adm[sibling, node] - adm[parent, sibling]
            )
        else:
            right = tree[node, 1]
            lens[node] = 0.25 * (
                adm[parent, left]
                + adm[sibling, right]
                + adm[parent, right]
                + adm[sibling, left]
            ) - 0.5 * (adm[parent, sibling] + adm[left, right])
    left, right = tree[0, 0], tree[0, 1]
    lens[0] = 0.5 * (adm[left, 0] + adm[right, 0] - adm[left, right])


def _ols_min_branch_d2(
    double[::1] lens,
    double[:, ::1] ad2,
    double[:, ::1] adk,
    Py_ssize_t[:, ::1] tree,
    Py_ssize_t[::1] preodr,
):
    r"""Find the branch with the minimum length change after inserting a new taxon.

    It returns the node at the lower end of the branch.

    Implemented according to Eq. 7 of Desper and Gascuel (2002).

    IMPORTANT NOTE: Should there are ties (which are common), this function returns
    the first minimum branch seen during preorder traversal. This behavior resembles
    FastME. Alternatively, one can return the first minimum branch by the order of
    addition using `return lens[:n].argmin()`, which could produce different results
    at the presence of ties. It must be noted that all candidates in a tie are equally
    optimal in the current iteration of the greedy algorithm.

    """
    cdef Py_ssize_t i
    cdef Py_ssize_t node, parent, sibling, size, p_size, s_size
    cdef double L, numerator, lambda_0, lambda_1

    cdef Py_ssize_t min_node = 0
    cdef double min_len = 0
    cdef Py_ssize_t m = tree[0, 4] + 1
    cdef Py_ssize_t n = 2 * m - 3

    lens[min_node] = min_len

    # Traverse tree in preorder and calculate the length change of each branch from its
    # previous branch.
    for i in range(1, n):
        node = preodr[i]
        parent = tree[node, 2]
        sibling = tree[node, 3]
        size = tree[node, 4]
        p_size = m - tree[parent, 4]
        s_size = tree[sibling, 4]

        numerator = s_size + size * p_size
        lambda_0 = numerator / ((s_size + size) * (p_size + 1))
        lambda_1 = numerator / ((s_size + p_size) * (size + 1))

        # factor 0.5 is omitted
        lens[node] = L = lens[parent] + (
            (lambda_0 - lambda_1) * (adk[sibling, 0] + ad2[node, 0])
            + (lambda_1 - 1) * (ad2[sibling, 1] + adk[parent, 1])
            + (1 - lambda_0) * (ad2[sibling, 0] + adk[node, 0])
        )

        if L < min_len:
            min_len, min_node = L, node

    return min_node


def _bal_min_branch(
    double[::1] lens,
    double[:, ::1] adm,
    double[:, ::1] adk,
    Py_ssize_t[:, ::1] tree,
    Py_ssize_t[::1] preodr,
):
    """Find the branch with the minimum length change after inserting a new taxon.

    This function resembles :func:`_ols_min_branch_d2` but it 1) uses the balanced
    framework and 2) calculates based on the entire matrix. See also the note of the
    latter.

    Implemented according to Eq. 10 of Desper and Gascuel (2002).

    """
    cdef Py_ssize_t i
    cdef Py_ssize_t node, parent, sibling
    cdef double L

    cdef Py_ssize_t min_node = 0
    cdef double min_len = 0

    lens[min_node] = min_len
    for i in range(1, 2 * tree[0, 4] - 1):
        node = preodr[i]
        parent = tree[node, 2]
        sibling = tree[node, 3]

        # factor 0.25 is omitted
        lens[node] = L = lens[parent] + (
            adm[sibling, parent] + adk[node, 0] - adm[sibling, node] - adk[parent, 1]
        )
        if L < min_len:
            min_len, min_node = L, node
    return min_node


def _avgdist_d2_insert(
    double[:, ::1] ad2,
    Py_ssize_t target,
    double[:, ::1] adk,
    Py_ssize_t[:, ::1] tree,
    Py_ssize_t[::1] preodr,
):
    r"""Update average distances between distant-2 subtrees after taxon insertion.

    This function will update ad2, a float array of (n, 2) representing pairwise
    distances between all distant-2 subtrees in the tree. Here, `distant-2 subtrees`
    refer to subtrees that are two branches away from each other. Specifically, there
    are two scenarios:

    - Column 0: Distance between the lower subtree of the current node and the upper
      subtree of its parent.
    - Column 1: Distance between the lower subtree of the current node and the lower
      subtree of its sibling.

    This function assumes that the taxon will be inserted into the branch connecting
    the target node and its parent. After insertion, the taxon will become the sibling
    of the target.

                                    parent
             parent                  /  \
             /  \        =>       link  sibling
        target  sibling           /  \
                             target  taxon (k)

    This function should be executed *before* calling :func:`_insert_taxon`, which will
    mutate the tree.

    Implemented according to Eq. 8 of Desper and Gascuel (2002).

    """
    cdef Py_ssize_t i, ii
    cdef Py_ssize_t node, left, right, parent, sibling, size, curr, p_size, size_1

    # dimensions and positions
    cdef Py_ssize_t m = tree[0, 4] + 1
    cdef Py_ssize_t n = 2 * m - 3
    cdef Py_ssize_t link = n
    cdef Py_ssize_t tip = n + 1

    ###### Special case: insert into the root branch. ######

    if target == 0:
        # k (lower) to root (parent, upper): pre-calculated
        ad2[tip, 0] = adk[0, 1]

        # k (lower) to link (sibling, lower): equals to k to root (lower).
        ad2[link, 1] = ad2[tip, 1] = adk[0, 0]

        # Link (lower) to root (parent, upper): de novo calculation according to the
        # equation in A4.1(b). It is basically the distance between the upper and lower
        # subtrees of the root itself.
        left, right = tree[0, 0], tree[0, 1]
        ad2[link, 0] = (
            tree[left, 4] * ad2[left, 0] + tree[right, 4] * ad2[right, 0]
        ) / tree[0, 4]

        # Calculate all node (lower) to parent (upper, containing k) distances. These
        # parents include the new link.
        for node in range(1, n):
            p_size = m - tree[tree[node, 2], 4]
            ad2[node, 0] = (adk[node, 0] + ad2[node, 0] * p_size) / (p_size + 1)

        return

    ###### Regular case: insert into any other branch. ######

    parent, sibling, size = tree[target, 2], tree[target, 3], tree[target, 4]
    size_1 = size + 1

    # Temporarily copy the distances of target to link (will edit later).
    ad2[link, 0] = ad2[target, 0]
    ad2[link, 1] = ad2[target, 1]

    # Distance between k (lower) and link (parent, upper) equals to that between k and
    # the upper subtree of target.
    ad2[tip, 0] = adk[target, 1]

    # Distance between target (lower) and link (parent, upper) needs to be calculated
    # using the equation in A4.1(c). Basically, it is the distance between the lower
    # and upper subtrees of the same target.
    ad2[target, 0] = (
        tree[sibling, 4] * ad2[target, 1] + (m - tree[parent, 4]) * ad2[target, 0]
    ) / (m - size)

    # Transfer the pre-calculated distance between target (lower) and k (sibling,
    # lower).
    ad2[target, 1] = ad2[tip, 1] = adk[target, 0]

    # Within the clade below target, calculate the distance between each node (lower)
    # and its parent (upper, containing k).
    ii = tree[target, 6]
    for i in range(ii + 1, ii + size * 2 - 1):
        node = preodr[i]
        p_size = m - tree[tree[node, 2], 4]
        ad2[node, 0] = (adk[node, 0] + ad2[node, 0] * p_size) / (p_size + 1)

    # Iterate over the ancestors of target, starting from link and ending at root.
    curr = link
    while curr:
        # Calculate the distance between each pair of lower (containing k) and upper
        # ancestors.
        ad2[curr, 0] = (adk[parent, 1] + ad2[curr, 0] * size) / size_1

        # Calculate the distance between each ancestor (lower, containing k) and its
        # sibling (lower).
        ad2[curr, 1] = ad2[sibling, 1] = (
            adk[sibling, 0] + ad2[curr, 1] * size
        ) / size_1

        # Within the clade below each sibling, calculate the distance between each node
        # (lower) and its parent (upper, containing k).
        ii = tree[sibling, 6]
        for i in range(ii + 1, ii + tree[sibling, 4] * 2 - 1):
            node = preodr[i]
            p_size = m - tree[tree[node, 2], 4]
            ad2[node, 0] = (adk[node, 0] + ad2[node, 0] * p_size) / (p_size + 1)

        curr = parent
        parent, sibling, size = tree[curr, 2], tree[curr, 3], tree[curr, 4]
        size_1 = size + 1


def _bal_avgdist_insert(
    double[:, ::1] adm,
    Py_ssize_t target,
    double[:, ::1] adk,
    Py_ssize_t[:, ::1] tree,
    Py_ssize_t[::1] preodr,
    Py_ssize_t[::1] postodr,
    double[::1] powers,
    Py_ssize_t[::1] stack,
):
    r"""Update balanced average distance matrix after taxon insertion.

    This function resembles :func:`_avgdist_d2_insert` but it 1) uses the balanced
    framework and 2) updates the entire matrix. The latter makes it the dominant term
    of the entire algorithm.

    Two additional parameters are provided: `powers` is a pre-calculated array of
    2^(-l) powers (l is the depth difference between two nodes). `stack` is an
    integer array to store ancestral nodes of target.

    """
    cdef Py_ssize_t i, j, ii, jj, anc_i
    cdef Py_ssize_t parent, sibling, depth
    cdef Py_ssize_t curr, anc, cousin, depth_1, depth_diff
    cdef Py_ssize_t a, b
    cdef double power, diff

    # dimensions and positions
    cdef Py_ssize_t m = tree[0, 4] + 1
    cdef Py_ssize_t n = 2 * m - 3
    cdef Py_ssize_t link = n
    cdef Py_ssize_t tip = n + 1

    ###### Special case: insert into the root branch. ######

    if target == 0:
        # Transfer distance between k and root (upper).
        adm[0, tip] = adm[tip, 0] = adk[0, 1]

        # k to link: equals to k to root (lower).
        adm[link, tip] = adm[tip, link] = adk[0, 0]

        # Root to link: de novo calculation according to the equation in A4.1(b). It is
        # basically the distance between the upper and lower subtrees of the root.
        a1, a2 = tree[0, 0], tree[0, 1]
        adm[0, link] = adm[link, 0] = 0.5 * (adm[a1, 0] + adm[a2, 0])

        # Iterate over all nodes but the root.
        for a in range(1, n):

            # Transfer distances between the node (lower) and k.
            adm[a, tip] = adm[tip, a] = adk[a, 0]
            
            # Calculate the distance between the node (lower) and link (upper, with two
            # taxa (0 and k) added) using Eq. 8.
            adm[a, link] = adm[link, a] = 0.5 * (adk[a, 0] + adm[a, 0])

            # Calculate the distances between the node (upper, containing k) and each
            # of its descendant (lower).
            ii = tree[a, 7]
            power = powers[tree[a, 5] + 1]
            for i in range(ii - tree[a, 4] * 2 + 2, ii):
                b = postodr[i]
                adm[a, b] = adm[b, a] = adm[a, b] + power * (adk[b, 0] - adm[0, b])

        return

    ###### Regular case: insert into any other branch. ######

    parent, sibling, depth = tree[target, 2], tree[target, 3], tree[target, 5]
    depth_1 = depth + 1

    ### Step 1: Distances within the clade below target. ###

    # Locate the clade below target (including target)
    ii = tree[target, 7]
    for i in range(ii - tree[target, 4] * 2 + 2, ii + 1):
        a = postodr[i]

        # Transfer pre-calculated distance between k (lower) and any node within the clade
        # (including target, lower).
        adm[a, tip] = adm[tip, a] = adk[a, 0]

        # Distance from any descendant (lower) to link (upper) equals to that to target.
        # (The last assignment: target to link is incorrect. It will be fixed below.)
        adm[a, link] = adm[link, a] = adm[a, target]

        # Within the clade, find all ancestor (a) - descendant (b) pairs, and calculate the
        # distance between the upper subtree of a (containing k) and the lower subtree of b.
        jj = tree[a, 7]
        power = powers[tree[a, 5] - depth + 1]
        for j in range(jj - tree[a, 4] * 2 + 2, jj):
            b = postodr[j]
            adm[a, b] = adm[b, a] = adm[a, b] + power * (adk[b, 0] - adm[target, b])

    ### Step 2: Distances around the insertion point. ###

    # Distance between k (lower) and link (upper) equals to that between k and the
    # upper subtree of target.
    adm[tip, link] = adm[link, tip] = adk[target, 1]

    # Distance between target (lower) and link (upper) needs to be calculated using the
    # equation in A4.1(c). Basically, it is the distance between the lower and upper
    # subtrees of the same target.
    adm[target, link] = adm[link, target] = 0.5 * (
        adm[target, sibling] + adm[target, parent]
    )

    ### Step 3: Distances among nodes outside the clade. ###

    # Iterate over ancestors of target in ascending order.
    anc_i = 0
    curr = target
    while curr:
        stack[anc_i] = anc = tree[curr, 2]
        depth_diff = depth - 2 * tree[anc, 5]

        # Transfer the pre-calculated distance between k and the ancestor (upper).
        adm[anc, tip] = adm[tip, anc] = adk[anc, 1]

        # Calculate the distance between link (lower, containing k) and the ancestor
        # (upper).
        adm[anc, link] = adm[link, anc] = 0.5 * (adk[anc, 1] + adm[anc, target])

        # Calculate the distance between each previous ancestor (lower, containing k)
        # and the current ancestor (upper).
        diff = adk[anc, 1] - adm[target, anc]
        for i in range(anc_i):
            a = stack[i]
            adm[anc, a] = adm[a, anc] = adm[anc, a] + powers[
                depth_1 - tree[a, 5]
            ] * diff

        # Identify the cousin clade descending from the ancestor.
        cousin = tree[curr, 3]
        ii = tree[cousin, 7]
        for i in range(ii - tree[cousin, 4] * 2 + 2, ii + 1):
            a = postodr[i]

            # Transfer the pre-calculated distances between k and each descendant
            # (lower).
            adm[a, tip] = adm[tip, a] = adk[a, 0]

            # Calculate the distance between link (lower, containing k) and each
            # descendant (lower).
            adm[a, link] = adm[link, a] = 0.5 * (adk[a, 0] + adm[a, target])

            # Calculate the distance between each previous ancestor (lower, containing
            # k) and each descendant (lower).
            diff = adk[a, 0] - adm[a, target]
            for j in range(anc_i):
                b = stack[j]
                adm[a, b] = adm[b, a] = adm[a, b] + powers[depth_1 - tree[b, 5]] * diff

            # Iterate over descendants of each member of the clade, and calculate the
            # distance between the former (upper, containing k) and the latter (lower).
            jj = tree[a, 7]
            power = powers[depth_diff + tree[a, 5]]
            for j in range(jj - tree[a, 4] * 2 + 2, jj):
                b = postodr[j]
                adm[a, b] = adm[b, a] = adm[a, b] + power * (
                    adk[b, 0] - adm[b, target]
                )

        curr = anc
        anc_i += 1


def _bal_avgdist_insert_p(
    double[:, ::1] adm,
    Py_ssize_t target,
    double[:, ::1] adk,
    Py_ssize_t[:, ::1] tree,
    Py_ssize_t[::1] preodr,
    Py_ssize_t[::1] postodr,
    double[::1] powers,
    Py_ssize_t[::1] stack,
):
    r"""Update balanced average distance matrix after taxon insertion.

    This function is the parallel version of :func:`_bal_avgdist_insert`.

    """
    cdef Py_ssize_t i, j, ii, jj, anc_i
    cdef Py_ssize_t parent, sibling, depth
    cdef Py_ssize_t curr, anc, cousin, depth_1, depth_diff
    cdef Py_ssize_t a, b
    cdef double power, diff

    # number of operations per iteration
    cdef int ops

    # dimensions and positions
    cdef Py_ssize_t m = tree[0, 4] + 1
    cdef Py_ssize_t n = 2 * m - 3
    cdef Py_ssize_t link = n
    cdef Py_ssize_t tip = n + 1

    ###### Special case: insert into the root branch. ######

    if target == 0:
        # Transfer distance between k and root (upper).
        adm[0, tip] = adm[tip, 0] = adk[0, 1]

        # k to link: equals to k to root (lower).
        adm[link, tip] = adm[tip, link] = adk[0, 0]

        # Root to link: de novo calculation according to the equation in A4.1(b). It is
        # basically the distance between the upper and lower subtrees of the root.
        a1, a2 = tree[0, 0], tree[0, 1]
        adm[0, link] = adm[link, 0] = 0.5 * (adm[a1, 0] + adm[a2, 0])

        # Iterate over all nodes but the root.
        for i in prange(
            n - 1, nogil=True, schedule="dynamic", chunksize=max(1, CHUNKSIZE // n)
        ):
            a = postodr[i]

            # Transfer distances between the node (lower) and k.
            adm[a, tip] = adm[tip, a] = adk[a, 0]

            # Calculate the distance between the node (lower) and link (upper, with two
            # taxa (0 and k) added) using Eq. 8.
            adm[a, link] = adm[link, a] = 0.5 * (adk[a, 0] + adm[a, 0])

            # Calculate the distances between the node (upper, containing k) and each
            # of its descendant (lower).
            ii = tree[a, 7]
            power = powers[tree[a, 5] + 1]
            for i in range(ii - tree[a, 4] * 2 + 2, ii):
                b = postodr[i]
                adm[a, b] = adm[b, a] = adm[a, b] + power * (adk[b, 0] - adm[0, b])

        return

    ###### Regular case: insert into any other branch. ######

    parent, sibling, depth = tree[target, 2], tree[target, 3], tree[target, 5]
    depth_1 = depth + 1

    ### Step 1: Distances around the insertion point. ###

    # Distance between k (lower) and link (upper) equals to that between k and the
    # upper subtree of target.
    adm[tip, link] = adm[link, tip] = adk[target, 1]

    # Distance between target (lower) and link (upper) needs to be calculated using the
    # equation in A4.1(c). Basically, it is the distance between the lower and upper
    # subtrees of the same target.
    adm[target, link] = adm[link, target] = 0.5 * (
        adm[target, sibling] + adm[target, parent]
    )

    # Transfer pre-calculated distance between target (lower) and k (lower).
    adm[target, tip] = adm[tip, target] = adk[target, 0]

    ### Step 2: Distances within the clade below target. ###

    # Locate the clade below target (excluding target)
    ii = tree[target, 7]
    ops = tree[target, 4] * 2 - 1
    for i in prange(
        ii - ops + 1, ii, nogil=True, schedule="dynamic", chunksize=max(1, CHUNKSIZE // ops)
    ):
        a = postodr[i]

        # Transfer pre-calculated distance between k (lower) and any node within the
        # clade (lower).
        adm[a, tip] = adm[tip, a] = adk[a, 0]

        # Distance from any descendant (lower) to link (upper) equals to that to
        # target.
        adm[a, link] = adm[link, a] = adm[a, target]

        # Within the clade, find all ancestor (a) - descendant (b) pairs, and calculate
        # the distance between the upper subtree of a (containing k) and the lower
        # subtree of b.
        jj = tree[a, 7]
        power = powers[tree[a, 5] - depth + 1]
        for j in range(jj - tree[a, 4] * 2 + 2, jj):
            b = postodr[j]
            adm[a, b] = adm[b, a] = adm[a, b] + power * (adk[b, 0] - adm[target, b])

    # Finally, calculate the distance between each node within the clade (lower) and
    # target (upper).
    for i in range(ii - ops + 1, ii):
        a = postodr[i]
        adm[a, target] = adm[target, a] = 0.5 * (adm[a, target] + adk[a, 0])

    ### Step 3: Distances among nodes outside the clade. ###

    # Iterate over ancestors of target in ascending order.
    anc_i = 0
    curr = target
    while curr:
        stack[anc_i] = anc = tree[curr, 2]
        depth_diff = depth - 2 * tree[anc, 5]

        # Transfer the pre-calculated distance between k and the ancestor (upper).
        adm[anc, tip] = adm[tip, anc] = adk[anc, 1]

        # Calculate the distance between link (lower, containing k) and the ancestor
        # (upper).
        adm[anc, link] = adm[link, anc] = 0.5 * (adk[anc, 1] + adm[anc, target])

        # Calculate the distance between each previous ancestor (lower, containing k)
        # and the current ancestor (upper).
        diff = adk[anc, 1] - adm[target, anc]
        for i in range(anc_i):
            a = stack[i]
            adm[anc, a] = adm[a, anc] = adm[anc, a] + powers[
                depth_1 - tree[a, 5]
            ] * diff

        # Identify the cousin clade descending from the ancestor.
        cousin = tree[curr, 3]
        ii = tree[cousin, 7]
        ops = tree[cousin, 4] * 2 - 1
        for i in prange(
            ii - ops + 1, ii + 1, nogil=True, schedule="dynamic", chunksize=max(1, CHUNKSIZE // ops)
        ):
            a = postodr[i]

            # Transfer the pre-calculated distances between k and each descendant
            # (lower).
            adm[a, tip] = adm[tip, a] = adk[a, 0]

            # Calculate the distance between link (lower, containing k) and each
            # descendant (lower).
            adm[a, link] = adm[link, a] = 0.5 * (adk[a, 0] + adm[a, target])

            # Calculate the distance between each previous ancestor (lower, containing
            # k) and each descendant (lower).
            diff = adk[a, 0] - adm[a, target]
            for j in range(anc_i):
                b = stack[j]
                adm[a, b] = adm[b, a] = adm[a, b] + powers[depth_1 - tree[b, 5]] * diff

            # Iterate over descendants of each member of the clade, and calculate the
            # distance between the former (upper, containing k) and the latter (lower).
            jj = tree[a, 7]
            power = powers[depth_diff + tree[a, 5]]
            for j in range(jj - tree[a, 4] * 2 + 2, jj):
                b = postodr[j]
                adm[a, b] = adm[b, a] = adm[a, b] + power * (
                    adk[b, 0] - adm[b, target]
                )

        curr = anc
        anc_i += 1


def _insert_taxon(
    Py_ssize_t taxon,
    Py_ssize_t target,
    Py_ssize_t[:, ::1] tree,
    Py_ssize_t[::1] preodr,
    Py_ssize_t[::1] postodr,
    bint use_depth=True,
):
    r"""Insert a taxon between a target node and its parent.

    For example, with the following local structure of the original tree:

          A
         / \
        B   C

    With target=B, this function inserts a taxon into the branch A-B. The structure
    becomes:

            A
           / \
        link  C
         / \
        B  taxon

    The link and taxon will be appended to the end of the tree array, but the pre- and
    postorders need to be muted such that new nodes can be inserted. Specifically:

        Preorder:  A - B - C => A - link - B - taxon - C
        Postorder: B - C - A => B - taxon - link - C - A

    A special case is that the taxon is inserted into the root branch (node=0). The
    tree becomes:

            A
           / \
        link taxon
         / \
        B   C

    The inserted taxon always becomes the right child.

    """
    # This function can be simplified by Python and NumPy APIs. Although I hoped that
    # NumPy vectorization can accelerate the code, especially the pre- and postorder
    # parts, the reality according to my tests is that cell-by-cell Cython code is
    # significantly faster than NumPy, and greatly reduces the overall runtime of the
    # entire algorithms. This effect is more obvious when the dataset is small, but
    # less so when it is large (but it is still there).
    #
    # The reason might be that when moving a block of elements within the same array,
    # NumPy needs to create a temporary array, but Cython can do the job in place.
    #
    # There might be a chance to re-consider NumPy (or even CuPy) API in the future.
    cdef Py_ssize_t left, right, parent, sibling, size, depth, pre_i, post_i
    cdef Py_ssize_t i, k, side, clade_n, pre_i_after, curr

    # determine tree dimensions
    # typically n = 2 * taxon - 3, but this function doesn't enforce this
    cdef Py_ssize_t m = tree[0, 4]
    cdef Py_ssize_t n = m * 2 - 1
    cdef Py_ssize_t link = n
    cdef Py_ssize_t tip = n + 1

    # Special case (root branch): taxon k becomes the sibling of all existing taxa
    # except for the root (taxon 0).
    if target == 0:
        # children
        left, right = tree[0, 0], tree[0, 1]
        tree[left, 2] = tree[right, 2] = link

        # root
        tree[0, 0] = link
        tree[0, 1] = tip
        tree[0, 4] = m + 1
        tree[0, 7] = n + 1

        # link
        tree[link, 0] = left
        tree[link, 1] = right
        tree[link, 2] = 0
        tree[link, 3] = tip
        tree[link, 4] = m
        tree[link, 5] = 1
        tree[link, 6] = 1
        tree[link, 7] = n - 1

        # tip
        tree[tip, 0] = 0
        tree[tip, 1] = taxon
        tree[tip, 2] = 0
        tree[tip, 3] = link
        tree[tip, 4] = 1
        tree[tip, 5] = 1
        tree[tip, 6] = n + 1
        tree[tip, 7] = n

        # entire tree depth + 1
        if use_depth:
            for i in range(1, n):
                tree[i, 5] += 1

        # preorder
        for i in range(n - 1, 0, -1):
            tree[i, 6] += 1
            preodr[i + 1] = preodr[i]
        preodr[1] = link
        preodr[n + 1] = tip

        # postorder
        postodr[n - 1] = link
        postodr[n] = tip
        postodr[n + 1] = 0

    # Regular case (any other branch): The link becomes the parent of the target node,
    # and child of its original parent. Taxon k becomes the sibling
    else:
        left = tree[target, 0]
        right = tree[target, 1]
        parent = tree[target, 2]
        sibling = tree[target, 3]
        size = tree[target, 4]
        depth = tree[target, 5]
        pre_i = tree[target, 6]
        post_i = tree[target, 7]

        side = int(tree[parent, 0] != target)
        tree[parent, side] = link
        tree[sibling, 3] = link
        tree[target, 2] = link
        tree[target, 3] = tip

        # preorder index of node after clade
        pre_i_after = pre_i + size * 2 - 1

        # link
        tree[link, 0] = target
        tree[link, 1] = tip
        tree[link, 2] = parent
        tree[link, 3] = sibling
        tree[link, 4] = size + 1
        tree[link, 5] = depth
        tree[link, 6] = pre_i
        tree[link, 7] = post_i + 2

        # tip
        tree[tip, 0] = 0
        tree[tip, 1] = taxon
        tree[tip, 2] = link
        tree[tip, 3] = target
        tree[tip, 4] = 1
        tree[tip, 5] = depth + 1
        tree[tip, 6] = pre_i_after + 1
        tree[tip, 7] = post_i + 1

        # clade depth +1
        if use_depth:
            for i in range(pre_i, pre_i_after):
                tree[preodr[i], 5] += 1

        # preorder shift: nodes after clade +2, tip inserted after clade, nodes within
        # clade +1, link inserted before clade
        for i in range(n - 1, pre_i_after - 1, -1):
            k = preodr[i]
            tree[k, 6] += 2
            preodr[i + 2] = k
        preodr[pre_i_after + 1] = tip

        for i in range(pre_i_after - 1, pre_i - 1, -1):
            k = preodr[i]
            tree[k, 6] += 1
            preodr[i + 1] = k
        preodr[pre_i] = link

        # postorder shift: all nodes after clade +2, tip and link inserted after clade
        for i in range(n - 1, post_i, -1):
            k = postodr[i]
            tree[k, 7] += 2
            postodr[i + 2] = k
        postodr[post_i + 2] = link
        postodr[post_i + 1] = tip

        # size +1 from link to root
        curr = link
        while curr:
            parent = tree[curr, 2]
            tree[parent, 4] += 1
            curr = parent


def _avgdist_swap(
    double[:, ::1] adm,
    Py_ssize_t target,
    Py_ssize_t side,
    Py_ssize_t[:, ::1] tree,
):
    r"""Update average distance matrix after branch swapping.

    This function will update adm, a float array of (n, n) representing pairwise
    distances between all subtrees in the tree.

    It assumes that one child of the target node on a certain side (left: 0, right: 1)
    exchanges position with the sibling of the target:

                 |                         |
               parent                    parent
               /   \                     /   \
           target  sibling    =>     target  child
            /  \                      /  \
        other  child              other  sibling

    Two categories of average distances will be updated:

        1. From target (upper) to any node within other and sibling.
        2. From target (lower) to any node within parent and child.

    Only target to itself doesn't need to be updated. All other nodes in the tree do.

    This function should be executed *after* calling :func:`_swap_branches`.

    Implemented according to A4.3(b) of Desper and Gascuel (2002).

    """
    cdef Py_ssize_t node

    # total numbers of taxa and nodes in the tree
    cdef Py_ssize_t m = tree[0, 4] + 1
    cdef Py_ssize_t n = 2 * m - 3

    # the two branches that were swapped
    cdef Py_ssize_t child = tree[target, 3]  # former child, now sibling
    cdef Py_ssize_t sibling = tree[target, side]  # former sibling, now child

    # the two branches that stay still
    cdef Py_ssize_t other = tree[target, 1 - side]  # the other child
    cdef Py_ssize_t parent = tree[target, 2]  # the parent

    cdef Py_ssize_t c_size = tree[child, 4]
    cdef Py_ssize_t o_size = tree[other, 4]
    cdef Py_ssize_t s_size = tree[sibling, 4]
    cdef Py_ssize_t p_size = m - tree[parent, 4]

    # Loop over all nodes except for target. These nodes can be divided into ones
    # within the clade below target (former sibling and other), and ones that are
    # outside the clade (former child and parent). Their distances to the target will
    # be updated separately.
    cdef Py_ssize_t start = tree[target, 6]
    cdef Py_ssize_t end = start + tree[target, 4] * 2 - 1

    # # 1) subtrees within (parent, child) vs sibling (lower) + other (lower)
    # for node in range(start):
    #     adm[node, target] = adm[target, node] = (
    #         s_size * adm[node, sibling] + o_size * adm[node, other]
    #     ) / (s_size + o_size)

    # # 2) # subtrees within (sibling, other) vs parent (upper) + child (lower)
    # for node in range(start + 1, end):
    #     adm[node, target] = adm[target, node] = (
    #         p_size * adm[node, parent] + c_size * adm[node, child]
    #     ) / (c_size + p_size)

    # # 3) same as 1)
    # for node in range(end, n):
    #     adm[node, target] = adm[target, node] = (
    #         s_size * adm[node, sibling] + o_size * adm[node, other]
    #     ) / (s_size + o_size)

    cdef double temp_val = adm[target, target]

    for node in range(n):
        # subtrees within (sibling, other) vs parent (upper) + child (lower)
        if start < tree[node, 6] < end:
            adm[node, target] = adm[target, node] = (
                p_size * adm[node, parent] + c_size * adm[node, child]
            ) / (c_size + p_size)

        # subtrees within (parent, child) vs sibling (lower) + other (lower)
        else:
            adm[node, target] = adm[target, node] = (
                s_size * adm[node, sibling] + o_size * adm[node, other]
            ) / (s_size + o_size)

    # this cell doesn't need to be filled and let's reset it
    adm[target, target] = temp_val


def _bal_avgdist_swap(
    double[:, ::1] adm,
    Py_ssize_t target,
    Py_ssize_t side,
    Py_ssize_t[:, ::1] tree,
    Py_ssize_t[::1] preodr,
    double[::1] powers,
    Py_ssize_t[::1] stack,
):
    r"""Update balanced average distance matrix after branch swapping.

    This function resembles :func:`_avgdist_swap`, but it uses a balanced framework.
    Specifically, it follows Eq. 18 and Appendix 5.3 of Desper and Gascuel (2002).

    This function is the dominant term of the entire BNNI algorithm.

    """
    cdef Py_ssize_t node, curr, before, after, cousin, a, b
    cdef Py_ssize_t i, j, ii, jj, anc_i, start, end
    cdef Py_ssize_t depth, depth_2, depth_diff
    cdef double power, diff

    cdef Py_ssize_t m = tree[0, 4] + 1
    cdef Py_ssize_t n = 2 * m - 3
    cdef Py_ssize_t child = tree[target, 3]
    cdef Py_ssize_t sibling = tree[target, side]
    cdef Py_ssize_t other = tree[target, 1 - side]
    cdef Py_ssize_t parent = tree[target, 2]

    # Step 1: Update distances between subtrees within each of the four clades (Eq. 18
    # and A5.3(a)).

    # 1.1: Work on the three downward clades rooted at sibling, child, and other. For
    # each of them, identify their neighbor (i.e., the other clade on the same side of
    # the target branch) before and after swapping.
    cdef Py_ssize_t arr[9]
    arr[0], arr[1], arr[2] = sibling, parent, other
    arr[3], arr[4], arr[5] = child, other, parent
    arr[6], arr[7], arr[8] = other, child, sibling
    for i in range(0, 9, 3):
        curr, before, after = arr[i], arr[i + 1], arr[i + 2]
        depth = tree[curr, 5]
        depth_2 = depth - 2
        ii = tree[curr, 6]
        for i in range(ii, ii + tree[curr, 4] * 2 - 1):
            a = preodr[i]
            power = powers[tree[a, 5] - depth_2]
            jj = tree[a, 6]
            for j in range(jj + 1, jj + tree[a, 4] * 2 - 1):
                b = preodr[j]
                adm[a, b] = adm[b, a] = adm[a, b] - power * (
                    adm[b, before] - adm[b, after]
                )

    # 1.2: Work on the upward clade rooted at the parent node. The process is similar
    # to that in `_bal_avgdist_insert`. Basically, we will take care three categories
    # of distances:
    # - previous ancestor vs. current ancestor
    # - previous & current ancestor vs. cousin or its descendant
    # - cousin or its descendant vs. descendant of the former
    before, after = sibling, child
    depth = tree[parent, 5]
    depth_2 = depth + 2

    # iterate over ancestors of the target node, starting from its parent
    anc_i = 0
    curr = parent
    while True:
        stack[anc_i] = curr
        depth_diff = depth_2 - 2 * tree[curr, 5] + 1

        # each previous ancestor vs. current ancestor
        diff = adm[curr, before] - adm[curr, after]
        for i in range(anc_i):
            a = stack[i]
            adm[curr, a] = adm[a, curr] = adm[curr, a] - powers[
                depth_2 - tree[a, 5]
            ] * diff

        if not curr:
            break
        anc_i += 1

        # iterate over nodes within each cousin clade (cousin and its descendants)
        cousin = tree[curr, 3]
        ii = tree[cousin, 6]
        for i in range(ii, ii + tree[cousin, 4] * 2 - 1):
            a = preodr[i]

            # each previous & current ancestor vs. current node
            diff = adm[a, before] - adm[a, after]
            for j in range(anc_i):
                b = stack[j]
                adm[a, b] = adm[b, a] = adm[a, b] - powers[depth_2 - tree[b, 5]] * diff

            # current node vs. each descendant
            jj = tree[a, 6]
            power = powers[depth_diff + tree[a, 5]]  #### trick
            for j in range(jj + 1, jj + tree[a, 4] * 2 - 1):
                b = preodr[j]
                adm[a, b] = adm[b, a] = adm[a, b] - power * (
                    adm[b, before] - adm[b, after]
                )

        curr = tree[curr, 2]

    # Step 2: Update distances between subtrees within each side of the target branch
    # and the other end of the branch (A5.3(b)).
    start = tree[target, 6]
    end = start + tree[target, 4] * 2 - 1
    for i in range(start):
        node = preodr[i]
        adm[node, target] = adm[target, node] = 0.5 * (
            adm[node, sibling] + adm[node, other]
        )
    for i in range(start + 1, end):
        node = preodr[i]
        adm[node, target] = adm[target, node] = 0.5 * (
            adm[node, parent] + adm[node, child]
        )
    for i in range(end, n):
        node = preodr[i]
        adm[node, target] = adm[target, node] = 0.5 * (
            adm[node, sibling] + adm[node, other]
        )

    # Step 3: Update the distance between the lower and upper subtrees of the target
    # node itself (A5.3(c)).

    # This step is omitted as we will not fill the diagonal of the matrix (i.e., the
    # distance between the lower and upper subtree of the same node). See `_avgdist_
    # matrix`.

    # adm[target, target] = 0.25 * (
    #     adm[parent, sibling]
    #     + adm[parent, other]
    #     + adm[child, sibling]
    #     + adm[child, other]
    # )


cdef void _ols_swap(
    double* L,
    Py_ssize_t* side,
    Py_ssize_t p_size,
    Py_ssize_t s_size,
    Py_ssize_t l_size,
    Py_ssize_t r_size,
    double ad_before,
    double ad_left,
    double ad_right,
) noexcept nogil:
    r"""Calculate the change in overall tree length after a given swap.

    In a given quartet AB|CD, the swap takes place between subtrees B and C. The
    decrease of overall branch length, delta L (larger is better), can be calculated
    according to Eq. 9 of Desper and Gascuel (2002):

        delta L = ((lambda - 1)(d(A, C) + d(B, D)) - (lambda' - 1)(d(A, B) + d(C, D))
            - (lambda - lambda')(d(A, D) + d(B, C))) / 2

    Where:

        lambda  = (|A||D| + |B||C|) / ((|A| + |B|)(|C| + |D|))
        lambda' = (|A||D| + |B||C|) / ((|A| + |C|)(|B| + |D|))

    This function calculates delta L for two scenarios: 1) swap between left child and
    sibling, and 2) swap between right child and sibling.

    These two calculations share some steps, which are exploited by the following code.

    Factor 0.5 (/2) is omitted. The output value is negated, in order to match Python's
    minheap data structure (smallest popped first).

    Then the smaller change if negative is retained, according to A4.2.

    """
    cdef Py_ssize_t size_0 = p_size * s_size + l_size * r_size
    cdef Py_ssize_t size_1 = p_size * r_size + s_size * l_size
    cdef Py_ssize_t size_2 = p_size * l_size + s_size * r_size

    cdef double ad_bl = ad_before - ad_left
    cdef double ad_br = ad_before - ad_right

    cdef double ad_lr_size = (ad_left - ad_right) / (size_1 + size_2)

    cdef double L1 = size_1 * (ad_br / (size_0 + size_1) - ad_lr_size) - ad_bl
    cdef double L2 = size_2 * (ad_bl / (size_0 + size_2) + ad_lr_size) - ad_br

    if L1 >= 0 and L2 >= 0:
        L[0] = 0
    elif L1 <= L2:
        L[0] = L1
        side[0] = 0
    else:
        L[0] = L2
        side[0] = 1


def _ols_all_swaps(
    double[::1] lens,
    Py_ssize_t[:, ::1] tree,
    double[:, ::1] adm,
):
    r"""Evaluate all possible swaps at all internal branches of a tree.

    Using an OLS framework.

    The results (length change, smaller is better) are saved to `lens`. The child side
    of each node that achieves this length change is saved to column 7 of the tree.

    Implemented according to A4.2(a) of Desper and Gascuel (2002).

    """
    cdef Py_ssize_t m = tree[0, 4] + 1
    cdef Py_ssize_t node, left, right, parent, sibling
    
    # root is zero
    lens[0] = 0

    # calculate on internal branches (nodes with children)
    for node in range(1, tree.shape[0]):
        if tree[node, 0]:
            left = tree[node, 0]
            right = tree[node, 1]
            parent = tree[node, 2]
            sibling = tree[node, 3]

            # calculate length change
            _ols_swap(
                &lens[node],
                &tree[node, 7],
                m - tree[parent, 4],
                tree[sibling, 4],
                tree[left, 4],
                tree[right, 4],
                adm[parent, sibling] + adm[left, right],
                adm[parent, left] + adm[sibling, right],
                adm[parent, right] + adm[sibling, left],
            )
        
        # tips are zero
        else:
            lens[node] = 0


def _ols_corner_swaps(
    Py_ssize_t target,
    list heap,
    double[::1] lens,
    Py_ssize_t[:, ::1] tree,
    double[:, ::1] adm,
):
    r"""Update swaps of the four corner branches of a swapped branch.

    Using an OLS framework.

    Specifically, the four corner branches concern all nodes illustrated below.

                   |
                   x
                  / \
             parent  x
               /   \
           target  sibling
            /  \      / \
        left  right  x   x
         / \    / \
        x   x  x   x

    Implemented according to A4.2(c) of Desper and Gascuel (2002).

    """
    cdef Py_ssize_t m = tree[0, 4] + 1
    cdef Py_ssize_t i, side
    cdef double gain
    cdef Py_ssize_t left, right, parent, sibling

    # update four corner branches if they are internal
    # 0: left, 1: right, 2: parent, 3: sibling
    # for 0, 1, 3, check left child; for 2, check parent
    for i in range(4):
        node = tree[target, i]
        if (node if i == 2 else tree[node, 0]):
            left = tree[node, 0]
            right = tree[node, 1]
            parent = tree[node, 2]
            sibling = tree[node, 3]

            # calculate length change
            _ols_swap(
                &lens[node],
                &tree[node, 7],
                m - tree[parent, 4],
                tree[sibling, 4],
                tree[left, 4],
                tree[right, 4],
                adm[parent, sibling] + adm[left, right],
                adm[parent, left] + adm[sibling, right],
                adm[parent, right] + adm[sibling, left],
            )

            # if length is reduced, push the swap into the heap
            if lens[node]:
                heappush(heap, (lens[node], node, tree[node, 7]))


def _bal_all_swaps(
    double[::1] gains,
    Py_ssize_t[::1] sides,
    Py_ssize_t[::1] nodes,
    double[:, ::1] adm,
    Py_ssize_t[:, ::1] tree,
):
    r"""Evaluate possible swaps at all internal branches of a tree.

    Using a balanced framework.

    Implemented according to Eq. 12 and A4.2(a) of Desper and Gascuel (2002).

        \delta L = ((d(A, B) + d(C, D)) - (d(A, C) + d(B, D))) / 4

    Larger and positive \delta L is favored.

    Factor 0.25 (/4) is omitted in the calculation.

    `lens` should be initiated with zeros, because root and tips are assigned zeros
    automatically and are skipped during the iteration.

    """
    cdef Py_ssize_t branch, node, left, right, parent, sibling
    cdef double L1, L2, Lcomm
    for branch in range(nodes.shape[0]):
        node = nodes[branch]
        left = tree[node, 0]
        if not left:
            continue
        right, parent, sibling = tree[node, 1], tree[node, 2], tree[node, 3]
        Lcomm = adm[parent, sibling] + adm[left, right]
        L1 = adm[parent, left] + adm[sibling, right]
        L2 = adm[parent, right] + adm[sibling, left]
        if L1 >= Lcomm and L2 >= Lcomm:
            gains[branch] = 0
        elif L1 <= L2:
            gains[branch], sides[branch] = Lcomm - L1, 0
        else:
            gains[branch], sides[branch] = Lcomm - L2, 1
